"""Represents a collection of data about how databases storing the same data
differ from each other.
"""

# TODO this class needs to be refactored, with attributes and methods
# simplified. The class is used in the compare pipeline, which has only
# been partially refactored since integrating into pdm_utils.
# So do NOT use this class for anything other than in the compare pipeline
# until it has been properly refactored.


# TODO refactor and test.
class DbCompareSummary:

    # Initialize all attributes:
    def __init__(self,matched_genomes_list):

        # Initialize all non-calculated attributes:
        self._matched_genomes_list = matched_genomes_list

        # Initialize all calculated attributes:

        # MySQL data
        # General genome data
        self._m_g_update_flag_tally = 0

        # Genome data checks
        self._m_gnms_with_nucleotide_errors_tally = 0
        self._m_gnms_with_translation_errors_tally = 0
        self._m_gnms_with_boundary_errors_tally = 0
        self._m_gnms_with_status_accession_error_tally = 0
        self._m_gnms_with_status_description_error_tally = 0

        # PhagesDB data
        # Genome data checks
        self._p_gnms_with_nucleotide_errors_tally = 0

        # GenBank data
        # Genome data checks
        self._g_gnms_with_nucleotide_errors_tally = 0
        self._g_gnms_with_translation_errors_tally = 0
        self._g_gnms_with_boundary_errors_tally = 0
        self._g_gnms_with_missing_locus_tags_tally = 0
        self._g_gnms_with_locus_tag_typos_tally = 0
        self._g_gnms_with_description_field_errors_tally = 0

        # MySQL-PhagesDB checks
        self._m_p_seq_mismatch_tally = 0
        self._m_p_seq_length_mismatch_tally = 0
        self._m_p_cluster_mismatch_tally = 0
        self._m_p_accession_mismatch_tally = 0
        self._m_p_host_mismatch_tally = 0

        # MySQL-GenBank checks
        self._m_g_seq_mismatch_tally = 0
        self._m_g_seq_length_mismatch_tally = 0
        self._m_g_header_phage_mismatch_tally = 0
        self._m_g_header_host_mismatch_tally = 0
        self._m_g_gnms_with_imperfectly_matched_ftrs_tally = 0
        self._m_g_gnms_with_unmatched_m_ftrs_tally = 0
        self._m_g_gnms_with_unmatched_g_ftrs_tally = 0
        self._m_g_gnms_with_different_descriptions_tally = 0
        self._m_g_gnms_with_different_translations_tally = 0
        self._m_g_gnms_with_author_errors_tally = 0

        # PhagesDB-GenBank checks
        self._p_g_seq_mismatch_tally = 0
        self._p_g_seq_length_mismatch_tally = 0


        # MySQL feature
        # Gene data checks
        self._m_translation_errors_tally = 0
        self._m_boundary_errors_tally = 0


        # GenBank feature
        # Gene data checks
        self._g_translation_errors_tally = 0
        self._g_boundary_errors_tally = 0
        self._g_missing_locus_tags_tally = 0
        self._g_locus_tag_typos_tally = 0
        self._g_description_field_errors_tally = 0

        # MySQL-GenBank checks
        self._m_g_different_descriptions_tally = 0
        self._m_g_different_start_sites_tally = 0
        self._m_g_different_translation_tally = 0
        self._m_g_unmatched_m_ftrs_tally = 0
        self._m_g_unmatched_g_ftrs_tally = 0


        # Calculate summary metrics
        self._m_total_gnms_analyzed = 0
        self._m_gnms_unmatched_to_p_tally = 0
        self._m_gnms_unmatched_to_g_tally = 0
        self._total_genomes_with_errors = 0


    # TODO refactor and test.
    def compute_summary(self, gnm_mysql, gnm_pdb, gnm_gbk):
        for gnms in self._matched_genomes_list:
            self.compute_matched_genomes_summary(gnms, gnm_mysql, gnm_pdb,
                                                 gnm_gbk)


    # TODO refactor and test.
    def  compute_matched_genomes_summary(self, gnms, gnm_mysql,
                                         gnm_pdb, gnm_gbk):
        """Check errors within matched genomes."""

        self._m_total_gnms_analyzed += 1

        if gnms._contains_errors:
            self._total_genomes_with_errors += 1

        m_gnm = gnms.m_genome
        p_gnm = gnms.p_genome
        g_gnm = gnms.g_genome

        if m_gnm.type == gnm_mysql:
            self._m_g_update_flag_tally += m_gnm.retrieve_record
            self.compute_mysql_gnm_summary(m_gnm)

        if p_gnm.type == gnm_pdb:
            self.compute_pdb_gnm_summary(p_gnm)
        else:
            self._m_gnms_unmatched_to_p_tally += 1

        if g_gnm.type == gnm_gbk:
            self.compute_gbk_gnm_summary(g_gnm)
        else:
            self._m_gnms_unmatched_to_g_tally += 1

        self.compute_mysql_pdb_summary(gnms)
        self.compute_mysql_gbk_summary(gnms)
        self.compute_pdb_gbk_summary(gnms)


    # TODO refactor and test.
    def  compute_mysql_gnm_summary(self, gnm):
        """Check errors within MySQL genome."""

        if gnm._nucleotide_errors:
            self._m_gnms_with_nucleotide_errors_tally += 1
        if gnm._cds_features_with_translation_error_tally > 0:
            self._m_gnms_with_translation_errors_tally += 1
            self._m_translation_errors_tally += \
                                gnm._cds_features_with_translation_error_tally
        if gnm._cds_features_boundary_error_tally > 0:
            self._m_gnms_with_boundary_errors_tally += 1
            self._m_boundary_errors_tally += \
                                gnm._cds_features_boundary_error_tally
        if gnm._status_accession_error:
            self._m_gnms_with_status_accession_error_tally += 1
        if gnm._status_description_error:
            self._m_gnms_with_status_description_error_tally += 1

    # TODO refactor and test.
    def  compute_pdb_gnm_summary(self, gnm):
        """Check errors within PhagesDB genome."""
        if gnm._nucleotide_errors:
            self._p_gnms_with_nucleotide_errors_tally += 1

    # TODO refactor and test.
    def  compute_gbk_gnm_summary(self, gnm):
        """Check errors within GenBank genome."""

        if gnm._nucleotide_errors:
            self._g_gnms_with_nucleotide_errors_tally += 1
        if gnm._cds_features_with_translation_error_tally > 0:
            self._g_gnms_with_translation_errors_tally += 1
            self._g_translation_errors_tally += \
                                gnm._cds_features_with_translation_error_tally
        if gnm._cds_features_boundary_error_tally > 0:
            self._g_gnms_with_boundary_errors_tally += 1
            self._g_boundary_errors_tally += \
                                    gnm._cds_features_boundary_error_tally
        if gnm._missing_locus_tags_tally > 0:
            self._g_gnms_with_missing_locus_tags_tally += 1
            self._g_missing_locus_tags_tally += gnm._missing_locus_tags_tally
        if gnm._locus_tag_typos_tally > 0:
            self._g_gnms_with_locus_tag_typos_tally += 1
            self._g_locus_tag_typos_tally += gnm._locus_tag_typos_tally
        if gnm._description_field_error_tally > 0:
            self._g_gnms_with_description_field_errors_tally += 1
            self._g_description_field_errors_tally += \
                                    gnm._description_field_error_tally


    # TODO refactor and test.
    def  compute_mysql_pdb_summary(self, gnms):
        """Check differences between MySQL and PhagesDB genomes."""

        if gnms._m_p_seq_mismatch:
            self._m_p_seq_mismatch_tally += 1
        if gnms._m_p_seq_length_mismatch:
            self._m_p_seq_length_mismatch_tally += 1
        if gnms._m_p_cluster_mismatch:
            self._m_p_cluster_mismatch_tally += 1
        if gnms._m_p_accession_mismatch:
            self._m_p_accession_mismatch_tally += 1
        if gnms._m_p_host_mismatch:
            self._m_p_host_mismatch_tally += 1

    # TODO refactor and test.
    def  compute_mysql_gbk_summary(self, gnms):
        """Check differences between MySQL and GenBank genomes."""

        if gnms._m_g_seq_mismatch:
            self._m_g_seq_mismatch_tally += 1
        if gnms._m_g_seq_length_mismatch:
            self._m_g_seq_length_mismatch_tally += 1
        if gnms._g_header_fields_name_mismatch:
            self._m_g_header_phage_mismatch_tally += 1
        if gnms._g_host_mismatch:
            self._m_g_header_host_mismatch_tally += 1
        if gnms._m_g_imperfect_matched_ftrs_tally > 0:
            self._m_g_gnms_with_imperfectly_matched_ftrs_tally += 1
            self._m_g_different_start_sites_tally += \
                                    gnms._m_g_imperfect_matched_ftrs_tally
        if gnms._m_ftrs_unmatched_in_g_tally > 0:
            self._m_g_gnms_with_unmatched_m_ftrs_tally += 1
            self._m_g_unmatched_m_ftrs_tally += \
                                    gnms._m_ftrs_unmatched_in_g_tally
        if gnms._g_ftrs_unmatched_in_m_tally > 0:
            self._m_g_gnms_with_unmatched_g_ftrs_tally += 1
            self._m_g_unmatched_g_ftrs_tally += \
                                    gnms._g_ftrs_unmatched_in_m_tally
        if gnms._m_g_different_descriptions_tally > 0:
            self._m_g_gnms_with_different_descriptions_tally += 1
            self._m_g_different_descriptions_tally += \
                                    gnms._m_g_different_descriptions_tally
        if gnms._m_g_different_translations_tally > 0:
            self._m_g_gnms_with_different_translations_tally += 1
            self._m_g_different_translation_tally += \
                                    gnms._m_g_different_translations_tally
        if gnms._m_g_author_error:
            self._m_g_gnms_with_author_errors_tally += 1

    # TODO refactor and test.
    def  compute_pdb_gbk_summary(self, gnms):
        """Check differences between PhagesDB and GenBank genomes."""

        if gnms._p_g_seq_mismatch:
            self._p_g_seq_mismatch_tally += 1
        if gnms._p_g_seq_length_mismatch:
            self._p_g_seq_length_mismatch_tally += 1



import pymysql as pms
from pdm_utils.functions import mysqldb
from pdm_utils.functions import basic

# TODO unittest entire class.
class RandomFieldUpdateHandler:
    def __init__(self, connection):
        """
        This object is a ticket validation/execution machine for
        performing any number of simple, single-field database updates
        on a MySQL database.
        :param connection: the connection to the MySQL db
        """
        self.connection = connection        # MySQL connection
        self.table = ""                        # database table
        self.table_valid = False            # assume table invalid until proven
        self.field = ""                        # field whose value will be changed
        self.field_valid = False            # assume field invalid until proven
        self.value = ""                        # replacement value
        self.key_name = ""                    # e.g. PhageID, GeneID
        self.key_name_valid = False            # assume key invalid until proven
        self.key_value = ""                    # e.g. Phrann, SEA_PHRANN_29
        self.key_value_valid = False        # assume invalid until proven
        self.valid_ticket = False            # can't execute ticket unless True

    def validate_table(self):
        """
        This function attempts to validate the table by simply querying
        for the table's description.
        :return:
        """
        try:
            cur = self.connection.cursor()
            cur.execute("DESCRIBE {}".format(self.table))
            cur.close()
            self.table_valid = True
        except pms.err.Error:
            self.table_valid = False
            print("\nInvalid table '{}'".format(self.table))
        return

    def validate_field(self):
        """
        This function attempts to validate the replacement field by
        checking whether it's on the list of fields in the indicated
        table.
        :return:
        """
        # Can't run this function without a valid table.
        if self.table_valid is False:
            return
        try:
            valid_fields = []
            cur = self.connection.cursor()
            cur.execute("DESCRIBE {}".format(self.table))
            fields = cur.fetchall()
            for field in fields:
                valid_fields.append(field[0])
            if self.field in valid_fields:
                self.field_valid = True
            else:
                self.field_valid = False
                print("\nInvalid replacement field '{}' for table "
                      "'{}'".format(self.field, self.table))
        except pms.err.Error as err:
            print("Error {}: {}".format(err.args[0], err.args[1]))
            self.table_valid = False
        return

    def validate_key_name(self):
        """
        This function attempts to validate the selection key by
        checking whether it's a field in the table marked as any kind
        of key.
        :return:
        """
        # Can't run this function without a valid table.
        if self.table_valid is False:
            return
        try:
            valid_keys = []
            cur = self.connection.cursor()
            cur.execute("DESCRIBE {}".format(self.table))
            fields = cur.fetchall()
            for field in fields:
                if field[2] != "":
                    valid_keys.append(field[0])
            if self.key_name in valid_keys:
                self.key_name_valid = True
            else:
                self.key_name_valid = False
                print("\nInvalid selection key '{}' for table '{}'".format(
                    self.key_name, self.table))
        except pms.err.Error as err:
            print("Error {}: {}".format(err.args[0], err.args[1]))
            self.key_name_valid = False
        return

    def validate_key_value(self):
        """
        This function attempts to validate the selection key's value by
        querying the database for the data associated with that key and
        value on the indicated table
        :return:
        """
        # Can't run this function without a valid table or selection key
        if self.table_valid is False:
            return
        elif self.key_name_valid is False:
            return
        try:
            cur = self.connection.cursor()
            cur.execute("SELECT * FROM {} WHERE {} = '{}'".format(
                self.table, self.key_name, self.key_value))
            tuples = cur.fetchall()
            if len(tuples) > 0:
                self.key_value_valid = True
            else:
                self.key_value_valid = False
                print("\nInvalid selection value '{}' for key '{}' in table "
                      "'{}'".format(self.key_value, self.key_name, self.table))
        except pms.err.Error as err:
            print ("Error {}: {}".format(err.args[0], err.args[1]))
            self.key_value_valid = False
        return

    def validate_ticket(self):
        """
        This function runs all 4 of the object's built-in ticket
        validation methods, and checks whether any of the ticket inputs
        were invalid.  If any are invalid, reject the ticket.  If none
        are invalid, accept the ticket.
        :return:
        """
        self.validate_table()
        self.validate_field()
        self.validate_key_name()
        self.validate_key_value()
        if False in [self.table_valid, self.field_valid, self.key_name_valid,
                self.key_value_valid]:
            self.valid_ticket = False
        else:
            self.valid_ticket = True
        return

    def execute_ticket(self):
        """
        This function checks whether the ticket is valid.  If it is not
        valid, the function returns with code 0, indicating failure to
        execute the ticket.  If the ticket is valid, request input from
        the user to verify that they actually want to proceed with the
        update they've proposed.  If response is in the affirmative,
        the ticket is executed.  Otherwise, indicate that this ticket
        will be skipped, and return 0 as the ticket was not executed.
        If an error is encountered during execution of the ticket,
        print error message and return 0.  If the ticket is executed
        without issue, return 1 indicating success.
        :return:
        """
        if self.valid_ticket is False:
            return 0
        try:
            command = mysqldb.create_update(self.table, self.field,
                        self.value, self.key_name, self.key_value)
            # print("\nCommand to execute:")
            # print(command)
            # prompt = "Do you wish to proceed? (y/n) "
            # result = basic.ask_yes_no(prompt=prompt, response_attempt=3)
            result = True
            if result == True:
                cur = self.connection.cursor()
                cur.execute(command)
                cur.execute("COMMIT")
                cur.close()
            else:
                print("Skipping this ticket...")
                return 0
        except pms.err.Error as err:
            print("Error {}: {}".format(err[0], err[1]))
            return 0
        return 1



"""Represents a structure to group three Genome objects and
perform comparisons between them to identify inconsistencies."""


# TODO this class needs to be refactored, with attributes and methods
# simplified. The class is used in the compare pipeline, which has only
# been partially refactored since integrating into pdm_utils.
# It relies on extra Genome object attributes that are NOT present
# in the base Genome class and that are added during compare pipeline.
# Eventually, GenomeTriad should be generalized and/or replaced with
# implementing multiple GenomePair objects to represent all pairwise
# comparisons.
# So do NOT use this class for anything other than in the compare pipeline
# until it has been properly refactored.

# Variables are prefixed to indicate genome type:
# GenBank =  "gbk", "g"
# MySQL = "mysql", "m"
# PhagesDB = "pdb", "p"

import re

from pdm_utils.functions import basic
from pdm_utils.classes import cdspair

# TODO refactor and test.
class GenomeTriad:
    """Stores three Genome objects."""
    # TODO: similar to GenomePair, but it has three Genome slots instead of two.
    # Many methods here have been improved elsewhere. Ultimately, this class
    # should be removed, and the compare pipeline should rely on GenomePair
    # objects. However, all GenomeTriad methods need to be accounted for,
    # and the compare pipeline would need to run specific checks in specific
    # GenomePair objects, based on the type of paired genomes
    # (MySQL-PhagesDB, MySQL-Genbank, etc.)

    # Initialize all attributes:
    def __init__(self):

        # Initialize all non-calculated attributes:
        self.m_genome = ""
        self.p_genome = ""
        self.g_genome = ""

        # MySQL and GenBank matched data comparison results
        self._m_g_seq_mismatch = False
        self._m_g_seq_length_mismatch = False
        self._g_header_fields_name_mismatch = False
        self._g_host_mismatch = False

        # List of MatchedCdsFeature objects
        self._m_g_perfect_matched_ftrs = []
        self._m_g_imperfect_matched_ftrs = []

        # List of Cds objects
        self._m_ftrs_unmatched_in_g = []
        self._g_ftrs_unmatched_in_m = []

        self._m_g_perfect_matched_ftrs_tally = 0
        self._m_g_imperfect_matched_ftrs_tally = 0
        self._m_ftrs_unmatched_in_g_tally = 0
        self._g_ftrs_unmatched_in_m_tally = 0
        self._m_g_different_descriptions_tally = 0
        self._m_g_different_translations_tally = 0
        self._m_g_author_error = False

        # MySQL and PhagesDB matched data comparison results
        self._m_p_seq_mismatch = False
        self._m_p_seq_length_mismatch = False
        self._m_p_host_mismatch = False
        self._m_p_accession_mismatch = False
        self._m_p_cluster_mismatch = False


        # PhagesDB and GenBank matched data comparison results
        self._p_g_seq_mismatch = False
        self._p_g_seq_length_mismatch = False

        # Total errors summary
        self._contains_errors = False
        self._total_number_genes_with_errors = 0


    # Define all attribute setters:


    # TODO refactor and test.
    def compare_mysql_gbk_genomes(self, gnm_mysql, gnm_gbk, cdspair_mysql_gbk):

        # verify that there is a MySQL and GenBank genome in
        # the matched genome object.
        m_gnm = self.m_genome
        g_gnm = self.g_genome
        m_cds_lst = m_gnm.cds_features

        if m_gnm.type == gnm_mysql and g_gnm.type == gnm_gbk:

            if m_gnm.seq != g_gnm.seq:
                self._m_g_seq_mismatch = True
            if m_gnm.length != g_gnm.length:
                self._m_g_seq_length_mismatch = True

            # Compare phage names
            pattern1 = re.compile("^" + m_gnm.name + "$")
            pattern2 = re.compile("^" + m_gnm.name)

            if (basic.find_expression(pattern2,g_gnm.description.split(" ")) == 0 or
                    basic.find_expression(pattern1,g_gnm.source.split(" ")) == 0 or
                    basic.find_expression(pattern1,g_gnm.organism.split(" ")) == 0 or
                    basic.find_expression(pattern1,g_gnm.source_feature_organism.split(" ")) == 0):

                self._g_header_fields_name_mismatch = True

            # Compare host_genus data
            search_host = m_gnm.host_genus
            if search_host == "Mycobacterium":
                search_host = search_host[:-3]
            pattern3 = re.compile("^" + search_host)

            if ((basic.find_expression(pattern3,g_gnm.description.split(" ")) == 0 or
                    basic.find_expression(pattern3,g_gnm.source.split(" ")) == 0 or
                    basic.find_expression(pattern3,g_gnm.organism.split(" ")) == 0 or
                    basic.find_expression(pattern3,g_gnm.source_feature_organism.split(" ")) == 0) or
                    (g_gnm.source_feature_host != "" and basic.find_expression(pattern3,g_gnm.source_feature_host.split(" ")) == 0) or
                    (g_gnm.source_feature_lab_host != "" and basic.find_expression(pattern3,g_gnm.source_feature_lab_host.split(" ")) == 0)):

                self._g_host_mismatch = True

            # Check author list for errors
            # For genomes with AnnotationAuthor = 1 (Hatfull), Graham is
            # expected to be an author.
            # For genomes with AnnotationAuthor = 0 (non-Hatfull/GenBank),
            # Graham is NOT expected to be an author.
            pattern5 = re.compile("hatfull")
            search_result = pattern5.search(g_gnm.authors.lower())
            if m_gnm.annotation_author == 1 and search_result == None:
                self._m_g_author_error = True
            elif m_gnm.annotation_author == 0 and search_result != None:
                self._m_g_author_error = True
            else:
                # Any other combination of MySQL and GenBank
                # author can be skipped.
                pass


            # Compare CDS features

            # First find all unique start-end-orientation CDS identifiers
            # for MySQL and GenBank genomes.
            m_start_end_strand_id_set = set()

            # All end_strand ids that are not unique
            m_start_end_strand_duplicate_id_set = set()

            for cds_ftr in m_cds_lst:
                if cds_ftr._start_end_strand_id not in m_start_end_strand_id_set:
                    m_start_end_strand_id_set.add(cds_ftr._start_end_strand_id)
                else:
                    m_start_end_strand_duplicate_id_set.add(cds_ftr._start_end_strand_id)
            # Remove the duplicate end_strand ids from the main id_set
            m_start_end_strand_id_set = \
                m_start_end_strand_id_set - m_start_end_strand_duplicate_id_set


            g_cds_list = g_gnm.cds_features
            g_start_end_strand_id_set = set()

            # All end_strand ids that are not unique
            g_start_end_strand_duplicate_id_set = set()
            for cds_ftr in g_cds_list:
                if cds_ftr._start_end_strand_id not in g_start_end_strand_id_set:
                    g_start_end_strand_id_set.add(cds_ftr._start_end_strand_id)

                else:
                    g_start_end_strand_duplicate_id_set.add(cds_ftr._start_end_strand_id)
            # Remove the duplicate end_strand ids from the main id_set
            g_start_end_strand_id_set = \
                g_start_end_strand_id_set - g_start_end_strand_duplicate_id_set

            # Create the perfect matched and unmatched sets
            m_unmatched_start_end_strand_id_set = \
                        m_start_end_strand_id_set - g_start_end_strand_id_set
            g_unmatched_start_end_strand_id_set = \
                        g_start_end_strand_id_set - m_start_end_strand_id_set
            perfect_matched_cds_id_set = \
                        m_start_end_strand_id_set & g_start_end_strand_id_set


            # From the unmatched sets, created second round of
            # end-orientation id sets MySQL end_strand data
            m_end_strand_id_set = set()

            # All end_strand ids that are not unique
            m_end_strand_duplicate_id_set = set()
            for cds_ftr in m_cds_lst:
                if cds_ftr._start_end_strand_id in m_unmatched_start_end_strand_id_set:
                    if cds_ftr._end_strand_id not in m_end_strand_id_set:
                        m_end_strand_id_set.add(cds_ftr._end_strand_id)
                    else:
                        m_end_strand_duplicate_id_set.add(cds_ftr._end_strand_id)

            # Remove the duplicate end_strand ids from the main id_set
            m_end_strand_id_set = \
                        m_end_strand_id_set - m_end_strand_duplicate_id_set


            g_end_strand_id_set = set()

            # All end_strand ids that are not unique
            g_end_strand_duplicates = set()
            for cds_ftr in g_cds_list:
                if cds_ftr._start_end_strand_id in g_unmatched_start_end_strand_id_set:
                    if cds_ftr._end_strand_id not in g_end_strand_id_set:
                        g_end_strand_id_set.add(cds_ftr._end_strand_id)
                    else:
                        g_end_strand_duplicates.add(cds_ftr._end_strand_id)

            # Remove the duplicate end_strand ids from the main id_set
            g_end_strand_id_set = \
                        g_end_strand_id_set - g_end_strand_duplicates


            # Create the imperfect matched set
            imperfect_matched_cds_id_set = \
                        m_end_strand_id_set & g_end_strand_id_set


            # Now go back through all CDS features and assign
            # to the appropriate dictionary or list.
            m_perfect_matched_cds_dict = {}
            m_imperfect_matched_cds_dict = {}
            m_unmatched_cds_list = []


            for cds_ftr in m_cds_lst:
                if cds_ftr._start_end_strand_id in perfect_matched_cds_id_set:
                    m_perfect_matched_cds_dict[cds_ftr._start_end_strand_id] = cds_ftr
                elif cds_ftr._end_strand_id in imperfect_matched_cds_id_set:
                    m_imperfect_matched_cds_dict[cds_ftr._end_strand_id] = cds_ftr
                else:
                    m_unmatched_cds_list.append(cds_ftr)


            g_perfect_matched_cds_dict = {}
            g_imperfect_matched_cds_dict = {}
            g_unmatched_cds_list = []

            for cds_ftr in g_cds_list:
                if cds_ftr._start_end_strand_id in perfect_matched_cds_id_set:
                    g_perfect_matched_cds_dict[cds_ftr._start_end_strand_id] = cds_ftr
                elif cds_ftr._end_strand_id in imperfect_matched_cds_id_set:
                    g_imperfect_matched_cds_dict[cds_ftr._end_strand_id] = cds_ftr
                else:
                    g_unmatched_cds_list.append(cds_ftr)


            # Create CdsPair objects
            # Compute matched gene errors and tallies
            # Perfectly matched features
            for start_end_strand_tup in perfect_matched_cds_id_set:

                matched_cds_object = cdspair.CdsPair()
                matched_cds_object.type = cdspair_mysql_gbk
                matched_cds_object.cds1 = \
                            m_perfect_matched_cds_dict[start_end_strand_tup]
                matched_cds_object.cds2 = \
                            g_perfect_matched_cds_dict[start_end_strand_tup]
                matched_cds_object.compare_cds()

                if matched_cds_object._total_errors > 0:
                    self._total_number_genes_with_errors += 1



                if matched_cds_object.different_translation:
                    self._m_g_different_translations_tally += 1
                if matched_cds_object.different_description:
                    self._m_g_different_descriptions_tally += 1
                self._m_g_perfect_matched_ftrs.append(matched_cds_object)


            # Imperfectly matched features
            for end_strand_tup in imperfect_matched_cds_id_set:

                matched_cds_object = cdspair.CdsPair()
                matched_cds_object.type = cdspair_mysql_gbk
                matched_cds_object.cds1 = \
                                m_imperfect_matched_cds_dict[end_strand_tup]
                matched_cds_object.cds2 = \
                                g_imperfect_matched_cds_dict[end_strand_tup]
                matched_cds_object.compare_cds()

                if matched_cds_object._total_errors > 0:
                    self._total_number_genes_with_errors += 1

                if matched_cds_object.different_description:
                    self._m_g_different_descriptions_tally += 1
                self._m_g_imperfect_matched_ftrs.append(matched_cds_object)


            # Compute unmatched error and gene total errors for
            # all unmatched features.
            for cds_ftr in m_unmatched_cds_list:
                cds_ftr._unmatched_error = True
                cds_ftr.check_for_errors()
                if cds_ftr._total_errors > 0:
                    self._total_number_genes_with_errors += 1

            for cds_ftr in g_unmatched_cds_list:
                cds_ftr._unmatched_error = True
                cds_ftr.check_for_errors()
                if cds_ftr._total_errors > 0:
                    self._total_number_genes_with_errors += 1


            # Set unmatched CDS lists
            self._m_ftrs_unmatched_in_g = m_unmatched_cds_list
            self._g_ftrs_unmatched_in_m = g_unmatched_cds_list

            # Now compute the number of features in each category
            self._m_g_perfect_matched_ftrs_tally = \
                                        len(self._m_g_perfect_matched_ftrs)
            self._m_g_imperfect_matched_ftrs_tally = \
                                        len(self._m_g_imperfect_matched_ftrs)
            self._m_ftrs_unmatched_in_g_tally = len(self._m_ftrs_unmatched_in_g)
            self._g_ftrs_unmatched_in_m_tally = len(self._g_ftrs_unmatched_in_m)


        # If there is no matching GenBank genome,
        # assign all MySQL genes to Unmatched.
        else:

            # Set unmatched CDS lists, but do NOT count them in unmatched tally.
            # Unmatched tally should reflect unmatched genes if there is
            # actually a matching GenBank genome.
            self._m_ftrs_unmatched_in_g = m_cds_lst

            # Now that all errors have been computed for each gene,
            # compute how many genes have errors
            # If there is no matching GenBank genome,
            # gene error tallies are only computed for the MySQL genome.
            m_gnm.compute_genes_with_errors_tally()
            self._total_number_genes_with_errors = m_gnm._genes_with_errors_tally


    # TODO refactor and test.
    def compare_mysql_phagesdb_genomes(self, gnm_mysql, gnm_pdb):

        # verify that there is a MySQL and PhagesDB genome
        # in the matched genome object.
        m_gnm = self.m_genome
        p_gnm = self.p_genome

        if m_gnm.type == gnm_mysql and p_gnm.type == gnm_pdb:

            if m_gnm.seq != p_gnm.seq:
                self._m_p_seq_mismatch = True
            if m_gnm.length != p_gnm.length:
                self._m_p_seq_length_mismatch = True
            if m_gnm.accession != p_gnm.accession:
                self._m_p_accession_mismatch = True
            if m_gnm.host_genus != p_gnm.host_genus:
                self._m_p_host_mismatch = True
            if m_gnm.cluster != p_gnm.cluster:
                self._m_p_cluster_mismatch = True


    # TODO refactor and test.
    def compare_phagesdb_gbk_genomes(self, gnm_pdb, gnm_gbk):

        # verify that there is a PhagesDB and GenBank genome
        # in the matched genome object.
        p_gnm = self.p_genome
        g_gnm = self.g_genome

        if p_gnm.type == gnm_pdb and g_gnm.type == gnm_gbk:

            if p_gnm.seq != g_gnm.seq:
                self._p_g_seq_mismatch = True
            if p_gnm.length != g_gnm.length:
                self._p_g_seq_length_mismatch = True



    # TODO refactor and test.
    def compute_total_genome_errors(self, gnm_gbk, gnm_pdb):
        m_gnm = self.m_genome
        g_gnm = self.g_genome
        p_gnm = self.p_genome

        if self._total_number_genes_with_errors > 0:
            self._contains_errors = True

        # MySQL genome
        if m_gnm._nucleotide_errors:
            self._contains_errors = True
        if m_gnm._status_description_error:
            self._contains_errors = True
        if m_gnm._status_accession_error:
            self._contains_errors = True

        # GenBank genome
        if g_gnm.type == gnm_gbk:
            if g_gnm._nucleotide_errors:
                self._contains_errors = True

        # PhagesDB genome
        if p_gnm.type == gnm_pdb:
            if p_gnm._nucleotide_errors:
                self._contains_errors = True

        # MySQL-GenBank
        if self._m_g_seq_mismatch:
            self._contains_errors = True
        if self._m_g_seq_length_mismatch:
            self._contains_errors = True
        if self._g_header_fields_name_mismatch:
            self._contains_errors = True
        if self._g_host_mismatch:
            self._contains_errors = True
        if self._m_g_imperfect_matched_ftrs_tally > 0:
            self._contains_errors = True
        if self._m_ftrs_unmatched_in_g_tally > 0:
            self._contains_errors = True
        if self._g_ftrs_unmatched_in_m_tally > 0:
            self._contains_errors = True
        if self._m_g_different_descriptions_tally > 0:
            self._contains_errors = True
        if self._m_g_different_translations_tally > 0:
            self._contains_errors = True
        if self._m_g_author_error:
            self._contains_errors = True

        # MySQL-PhagesDB
        if self._m_p_seq_mismatch:
            self._contains_errors = True
        if self._m_p_seq_length_mismatch:
            self._contains_errors = True
        if self._m_p_host_mismatch:
            self._contains_errors = True
        if self._m_p_accession_mismatch:
            self._contains_errors = True
        if self._m_p_cluster_mismatch:
            self._contains_errors = True

        # PhagesDB-GenBank
        if self._p_g_seq_mismatch:
            self._contains_errors = True
        if self._p_g_seq_length_mismatch:
            self._contains_errors = True



"""Represents a collection of data about a tmRNA feature that are commonly used
to maintain and update SEA-PHAGES phage genomics data.
"""

import re
from collections import OrderedDict

from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, FeatureLocation, CompoundLocation
from Bio.Alphabet import IUPAC

from pdm_utils.classes import evaluation
from pdm_utils.classes.aragornhandler import AragornHandler
from pdm_utils.functions import basic

# Extracts peptide tag from note field acid and anticodon from note field for Aragorn-determinate
# or tRNAscan-SE- determinate or indeterminate tRNAs
NOTE_STANDARD_REGEX = re.compile("Tag peptide:\s+([\w|*]*)")


class Tmrna:
    def __init__(self):
        """
        Constructor method for a tmRNA object.
        """
        # The only attribute a tmRNA really needs to have is a sequence
        self.seq = Seq("", IUPAC.ambiguous_dna)
        self.length = 0

        # Information about the tmRNA with respect to its parent genome
        self.genome_id = ""     # Identifier for the parent genome
        self.genome_length = -1
        self.start = -1         # Start coord in parent genome, 0-indexed
        self.stop = -1          # Stop coord in parent genome, 0-indexed
        self.coordinate_format = ""
        self.orientation = ""   # "F" or "R", etc. - relative to parent genome

        # MySQL tmRNAs will also need these:
        self.id = ""            # Identifier for this gene
        self.name = ""          # tRNA number in the parent genome
        self.peptide_tag = ""   # Degradation tag

        # Genbank-formatted flat files should also have these attributes:
        self.seqfeature = None  # BioPython SeqFeature object for this tRNA
        self.locus_tag = ""     # Gene ID comprised of PhageID and Gene name
        self._locus_tag_num = ""
        self._gene_num = ""
        self.parts = 0          # Number of regions this gene encompasses
        self.gene = ""          # Gene number parsed from feature
        self.note = ""          # Raw note field

        # Aragorn data
        self.aragorn_run = False
        self.aragorn_data = None

        # Useful for processing data from various sources:
        self.evaluations = list()
        self.type = ""
        self._start_stop_orient_id = tuple()
        self._end_orient_id = tuple()
        self._start_end_id = tuple()

    # TODO: create base feature class
    def set_locus_tag(self, tag="", delimiter="_", check_value=None):
        """
        Populate the locus_tag and parse the locus_tag number.
        :param tag: Input locus_tag data
        :type tag: str
        :param delimiter: Value used to split locus_tag data
        :type delimiter: str
        :param check_value: Genome name or other value that will be
        used to parse the locus_tag to identify the feature number
        :type check_value: str
        """
        # List of delimiters commonly found in locus tags. "_" is most common
        delims = ["_", "-", "."]

        # If tag is None, locus_tag and _locus_tag_num remain default ""
        if tag is None:
            return

        # Else, we'll set the tag directly, then decide how to parse number
        self.locus_tag = tag

        # If delimiter is None, choose based on what appears in locus_tag
        if delimiter is None:
            delimiter = basic.choose_most_common(tag, delims)

        # If no check_value was given, use self.genome_id
        if check_value is None:
            check_value = self.genome_id

        # Compile a search pattern and split the incoming tag on delimiter
        pattern = re.compile(check_value.lower())
        parts = tag.split(delimiter)

        # Attempt to parse locus tag number from the locus tag
        parsed_value = None
        for i in range(len(parts)):
            # Search each part of the locus tag for check_value
            search_result = pattern.search(parts[i].lower())
            if search_result is not None:
                if i == len(parts) - 1:
                    # If the check_value was found in last part of locus tag,
                    # value is probably everything after check_value
                    parsed_value = parts[i][search_result.end():]
                else:
                    # If the check_value was found anywhere else in locus tag,
                    # value is probably found in the next index
                    parsed_value = parts[-1]
                # No need to keep searching if we found a match
                break

        if parsed_value is None:
            # Last-ditch effort to find locus tag number
            parsed_value = parts[-1]

        # Remove generic 'gp' or 'orf' prefixes
        if parsed_value.lower().startswith("gp"):
            parsed_value = parsed_value[2:]
        elif parsed_value.lower().startswith("orf"):
            parsed_value = parsed_value[3:]
        else:
            pass

        # If the value we parsed is a digit, populate _locus_tag_num
        left, right = basic.split_string(parsed_value)
        if right != "":
            self._locus_tag_num = right
        else:
            self._locus_tag_num = parsed_value

    # TODO: create base feature class
    def set_name(self, value=None):
        """
        Set the feature name. Ideally, the name of the CDS will be an
        integer. This information can be stored in multiple fields in
        the GenBank-formatted flat file. The name is derived from one
        of several qualifiers.
        :param value: Indicates a value that should be used to
        directly set the name regardless of the 'gene' and
        '_locus_tag_num' attributes.
        :type value: str
        """
        # The CDS feature number may be stored in several places in the
        # gene, product, note, function, and locus_tag qualifiers of a
        # flat file, but it is unpredictable. Not all of these qualifiers
        # are always present.
        # 1. PECAAN Draft and new SEA-PHAGES Final:
        #    The 'gene' qualifier should be present and contain an integer.
        # 2. SEA-PHAGES Final in GenBank:
        #    The 'gene' qualifier may or may not be present, and
        #    it may or may not have an integer.
        #    The 'locus_tag' qualifier may or may not be present,
        #    and may or may not have an integer.
        # 3. Non-SEA-PHAGES in GenBank:
        #    The 'gene' qualifier may or may not be present, and
        #    it may or may not have an integer.
        #    The 'locus_tag' qualifier may or may not be present,
        #    and may or may not have an integer.
        if value is None:
            name = ""
            list1 = ["_locus_tag_num", "_gene_num"]

            # First see if any num attributes have a float.
            x = 0
            while name == "" and x < len(list1):
                value = getattr(self, list1[x])
                if basic.is_float(value):
                    name = value
                x += 1

            # Second see if any num attributes have non-empty values.
            # At this point, it's very unpredictable. Values could be like
            # 'terL' in the gene or like '10a' in the locus_tag.
            if name == "":
                list2 = ["gene"]
                list2.extend(list1)
                y = 0
                while name == "" and y < len(list2):
                    value = getattr(self, list2[y])
                    if value != "":
                        name = value
                    y += 1
        self.name = value

    # TODO: create base feature class
    def set_num(self, attr, description, delimiter=None, prefix_set=None):
        """
        Set a number attribute from a description.
        :param attr: Attribute to set the number.
        :type attr: str
        :param description: Description data from which to parse the number.
        :type description: str
        :param delimiter: Value used to split the description data.
        :type delimiter: str
        :param prefix_set: Valid possible delimiters in the description.
        :type prefix_set: set
        """
        # Used to set product_num, function_num, note_num, gene_num
        # Sometimes feature number is stored within the description
        # (e.g. 'gp10; terminase' or 'terminase; gp10')
        # List of delimiters found in description fields
        # ";" is probably the most common.
        delimiters = [";", ","]
        if delimiter is None:
            delimiter = basic.choose_most_common(description, delimiters)
        if prefix_set is None:
            prefix_set = {"gp", "orf", ""}

        # Iterate through the list of strings, and select the first
        # string that looks like a gene number.
        split_value = description.split(delimiter)
        num = ""
        x = 0
        while (num == "" and x < len(split_value)):
            string = split_value[x]
            string = string.strip()
            left, right = basic.split_string(string)
            # Possible returns:
            # 1. left is alpha and right is number or float (gp10 = gp, 10),
            # 2. string is alpha (terminase = terminase, ""),
            # 3. string is number (10 = "", 10).
            if left.lower() in prefix_set:
                num = right
            x += 1
        setattr(self, attr, num)

    def set_gene(self, value, delimiter=None, prefix_set=None):
        """Set the gene attribute.

        :param value: Gene data to parse. Also passed to set_num().
        :type value: str
        :param delimiter: Passed to set_num().
        :type delimiter: str
        :param prefix_set: Passed to set_num().
        :type prefix_set: set
        """
        value = value.strip()
        self.gene = value
        self._gene_num = value
        self.set_num("_gene_num", value,
                     delimiter=delimiter, prefix_set=prefix_set)

    # TODO: create base feature class - use this version of the method unless
    #  Travis objects, because it's more Pythonic.
    def get_begin_end(self):
        """
        Accesses feature coordinates in transcription begin-end format.
        :return: (begin, end)
        """
        # Get a copy of the orientation in fr_long format:
        orientation = basic.reformat_strand(self.orientation, "fr_long")

        if orientation == "forward":
            # Rightward transcribed gene
            begin, end = self.start, self.stop
        elif orientation == "reverse":
            # Leftward transcribed gene
            begin, end = self.stop, self.start
        else:
            # Unexpected orientation
            begin = end = -1

        return begin, end

    # TODO: create base feature class - use this version of the method unless
    #  Travis objects, because it's documented more cleanly and avoids using
    #  keyword 'format'.
    def set_orientation(self, value, fmt, case=False):
        """
        Set the orientation based on the indicated format.
        :param value: orientation value
        :type value: int or str
        :param fmt: how orientation should be formatted
        :type fmt: str
        :param case: whether to capitalize the first letter of
        orientation
        :type case: bool
        :return:
        """
        self.orientation = basic.reformat_strand(value, fmt, case)

    # TODO: create base feature class
    def set_location_id(self):
        """
        Create identifier tuples containing feature location data. For
        this method we only care about gene boundaries and will ignore
        any multi-part elements to the gene.
        :return:
        """
        # Tuple with ALL relevant information
        self._start_stop_orient_id = (self.start, self.stop, self.orientation)

        begin, end = self.get_begin_end()

        # If only end coordinate is used, we need to know what orientation the
        # gene is in - two genes could theoretically be transcribed in opposed
        # directions and have the same stop position
        self._end_orient_id = (end, self.orientation)

        # Or we can have transcription begin/end coordinates, in which case
        # the direction can be gleaned from these values
        self._start_end_id = (begin, end)

    # TODO: create base feature class - use this version of the method unless
    #  Travis objects, because it has slightly improved logic.
    def reformat_start_and_stop(self, fmt):
        """
        Convert existing start and stop coordinates to the indicated
        new format; also updates the coordinate format attribute to
        reflect any change.
        :param fmt: the new desired coordinate format
        :type fmt: str
        :return:
        """
        # Only need to do work if the new coordinate format differs from
        # the existing coordinate format
        if fmt != self.coordinate_format:
            self.start, self.stop = basic.reformat_coordinates(
                self.start, self.stop, self.coordinate_format, fmt)
            self.coordinate_format = fmt

    # TODO: create base feature class - fully equivalent to version in Cds,
    #  but documented differently.
    def set_nucleotide_length(self, use_seq=False):
        """
        Set the nucleotide length of this gene feature.
        :param use_seq: whether to use the Seq feature to calculate
        nucleotide length of this feature
        :type use_seq: bool
        :return:
        """
        if use_seq:
            self.length = len(self.seq)
        else:
            if self.coordinate_format == "0_half_open":
                # Python coordinate format
                self.length = self.stop - self.start
            elif self.coordinate_format == "1_closed":
                # Genbank coordinate format
                self.length = self.stop - self.start + 1
            else:
                # Unknown coordinate format
                self.length = -1

    # TODO: create base feature class - use this version of the method unless
    #  Travis objects, because the logic is slightly better and it raises an
    #  Error if something is wrong with given seq.
    def set_nucleotide_sequence(self, value=None, parent_genome_seq=None):
        """
        Set this feature's nucleotide sequence
        :param value: sequence
        :type value: str or Seq
        :param parent_genome_seq: parent genome sequence
        :type parent_genome_seq: Seq
        :raise: ValueError
        :return:
        """
        # If seq is given we'll try to use that
        if value is not None:
            # If seq is a BioPython Seq object, use it directly
            if isinstance(value, Seq):
                self.seq = value.upper()
            # If seq is a string, we can coerce it to a BioPython Seq object
            elif isinstance(value, str):
                self.seq = Seq(value.upper(), IUPAC.ambiguous_dna)
            # If seq is something else entirely, we cannot use it
            else:
                pass
                # raise ValueError(f"tRNA.seq cannot use type '{type(seq)}'")
        # If instead a parent_genome_seq is given, we'll try to use that
        elif parent_genome_seq is not None and self.seqfeature is not None:
            try:
                self.seq = self.seqfeature.extract(parent_genome_seq)
            # TODO: tighten exception clause, OR let the error pass to a higher
            #  level in the call stack
            except:
                # Leave as default
                pass
        # If neither is given, leave self.seq as default
        else:
            pass

    # TODO: create base feature class - fully equivalent to the version in Cds,
    #  but better documented.
    def set_seqfeature(self):
        """
        Create a SeqFeature object with which to populate the
        `seqfeature` attribute.
        :return:
        """
        # SeqFeature coordinates are 0-based half-open
        start, stop = basic.reformat_coordinates(
            self.start, self.stop, self.coordinate_format, "0_half_open")

        # SeqFeature orientation is (-1, 1) instead of ("R", "F")
        strand = basic.reformat_strand(self.orientation, "numeric")

        # Standard genes will have start < stop
        if self.start <= self.stop:
            self.seqfeature = SeqFeature(FeatureLocation(start, stop),
                                         strand=strand, type=self.type)
        # Wrap-around genes will have stop < start
        else:
            self.seqfeature = SeqFeature(CompoundLocation(
                [FeatureLocation(start, self.genome_length),
                 FeatureLocation(0, stop)]), strand=strand, type=self.type)
        # Add feature qualifiers
        self.seqfeature.qualifiers = self.get_qualifiers()

    def get_qualifiers(self):
        """
        Helper function that uses tRNA data to populate the qualifiers
        attribute of `seqfeature`.
        :return: qualifiers OrderedDict()
        """
        qualifiers = OrderedDict()
        qualifiers["gene"] = [self.name]
        if self.locus_tag != "":
            qualifiers["locus_tag"] = [self.locus_tag]
        if self.peptide_tag != "":
            qualifiers["note"] = [f"Tag peptide: {self.peptide_tag}"]

        return qualifiers

    def run_aragorn(self):
        """
        Uses an AragornHandler object to negotiate the flow of
        information between this object and Aragorn.
        :return:
        """
        if self.id is not None and self.id != "":
            identifier = self.id
        else:
            identifier = "aragorn"

        if len(self.seq) > 0:
            ah = AragornHandler(self.id, str(self.seq))
            ah.write_fasta()
            ah.run_aragorn(m=True, t=False)  # search (linear) self.seq for tmRNAs
            ah.read_output()
            ah.parse_tmrnas()
            if ah.tmrna_tally == 1:
                self.aragorn_data = ah.tmrnas[0]
            # else:
                # print(f"Aragorn found {ah.tmrna_tally} tmRNAs in this region.")
            self.aragorn_run = True
        else:
            print("Cannot run Aragorn on 0-length sequence.")

    def parse_peptide_tag(self):
        """
        Parse the `peptide_tag` attribute out of the note field.
        :return:
        """
        # For SEA annotations, the note should contain the peptide tag
        results = NOTE_STANDARD_REGEX.findall(self.note)
        if len(results) == 1:
            self.peptide_tag = results[0]

    # Evaluations
    # TODO: create base feature class - fully equivalent to version in Cds,
    #  but documented slightly differently.
    def set_eval(self, eval_id, definition, result, status):
        """
        Constructs and adds and Evaluation object to this feature's list of
        evaluations.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param definition: description of the evaluation
        :type definition: str
        :param result: description of the evaluation outcome
        :type result: str
        :param status: overall outcome of the evaluation
        :type status: str
        :return:
        """
        evl = evaluation.Evaluation(eval_id, definition, result, status)
        self.evaluations.append(evl)

    # TODO: create base feature class - nearly identical to Cds version,
    #  but exception clause is appropriately narrow.
    def check_attribute(self, attribute, check_set, expect=False, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """
        Checks whether the indicated feature attribute is present in
        the given check_set. Uses expect to determine whether the
        presence (or lack thereof) is an error, or correct.
        :param attribute: the gene feature attribute to evaluate
        :type attribute: str
        :param check_set: set of reverence values
        :type check_set: set
        :param expect: whether the attribute's value is expected to be
        in the reference set
        :type expect: bool
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        try:
            value = getattr(self, attribute)
        except AttributeError:
            value = None

        if value is not None:
            value_short = basic.truncate_value(str(value), 30, "...")
            result = f"The '{attribute}' value '{value_short}' is "

            # Returns a boolean
            outcome = basic.check_value_expected_in_set(
                value, check_set, expect)

            if outcome:
                result += "valid."
                status = success
            else:
                result += "not valid."
                status = fail
        else:
            result = f"'{attribute}' is not a valid attribute to be evaluated."
            status = "untested"

        definition = f"Check the value of the '{attribute}' attribute for " \
                     f"{self.locus_tag}({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO: create base feature class - Travis' version assumes value can't
    #  be None; if wrong, that method will fail because None can't be <=> x.
    def check_magnitude(self, attribute, expect, ref_value, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """
        Check that the magnitude of a numerical attribute meets
        expectations.
        :param attribute: the gene feature attribute to evaluate
        :type attribute: str
        :param expect: symbol designating direction of magnitude (>=<)
        :type expect: str
        :param ref_value: numerical value for comparison
        :type ref_value: int, float, datetime
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        try:
            value = getattr(self, attribute)
        except AttributeError:
            value = None

        if value is not None:
            result = f"The '{attribute}' value '{value}' is "
            if value > ref_value:
                compare = ">"
                result += "greater than "
            elif value == ref_value:
                compare = "="
                result += "equal to "
            else:
                compare = "<"
                result += "less than "
            result += f"{ref_value}, which is "
            if compare == expect:
                result += "expected."
                status = success
            else:
                result += "not expected."
                status = fail
        else:
            result = f"'{attribute}' is not a valid attribute to be evaluated."
            status = "untested"

        definition = f"Check the magnitude of the '{attribute}' attribute " \
                     f"for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO: create base feature class - this version avoids overloading
    #  built-in format()
    def check_orientation(self, fmt="fr_short", case=True, eval_id=None,
                          success="correct", fail="error", eval_def=None):
        """
        Check that the orientation is set appropriately.
        :param fmt: indicates how coordinates should be formatted
        :type fmt: str
        :param case: indicates whether orientation data should be cased
        :type case: bool
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        expect = basic.reformat_strand(self.orientation, fmt, case)
        result = f"The orientation is {self.orientation}, and it is formatted "

        if self.orientation == expect:
            result += "correctly."
            status = success
        else:
            result += "incorrectly."
            status = fail

        definition = f"Check if the orientation is set correctly for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_orientation_correct(self, fmt="fr_short", case=True,
                                  eval_id=None, success="correct",
                                  fail="error", eval_def=None):
        """
        Check that the orientation agrees with the Aragorn and/or
        tRNAscan-SE predicted orientation. If Aragorn/tRNAscan-SE
        report a forward orientation, it means they agree with the
        annotated orientation. If they report reverse orientation,
        they think the annotation is backwards.
        :param fmt: indicates how coordinates should be formatted
        :type fmt: str
        :param case: indicates whether orientation data should be cased
        :type case: bool
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        # Make sure self.orientation is "F"/"R"
        orientation = basic.reformat_strand(self.orientation, fmt, case)

        # If Aragorn predicts a tmRNA, get orientation in "F"/"R" format
        if self.aragorn_data is not None:
            aragorn_orient = basic.reformat_strand(
                self.aragorn_data["Orientation"], fmt, case)
        else:
            aragorn_orient = None

        result = f"The annotated orientation '{orientation}' "

        # If only Aragorn predicts a tRNA
        if aragorn_orient is not None:
            # If Aragorn predicts "F" it agrees with annotated orientation
            if aragorn_orient == "F":
                result += "matches the Aragorn prediction."
                status = success
            else:
                result += "is backwards relative to the Aragorn prediction."
                status = fail
        # If Aragorn doesn't predict a tmRNA, we presume annotation is wrong.
        else:
            result += "is at odds with expectations (NO tmRNA HERE)."
            status = fail

        definition = f"Check whether the annotated orientation for {self.locus_tag} ({self.id})" \
                     f" matches the Aragorn prediction."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO: create base feature class - I think this is exactly the same as
    #  in Cds
    def check_locus_tag_structure(self, check_value=None, only_typo=False,
                                  prefix_set=set(), case=True, eval_id=None,
                                  success="correct", fail="error", eval_def=None):
        """Check if the locus_tag is structured correctly.

        :param check_value:
            Indicates the genome id that is expected to be present.
            If None, the 'genome_id' parameter is used.
        :type check_value: str
        :param only_typo:
            Indicates if only the genome id spelling should be evaluated.
        :type only_typo: bool
        :param prefix_set:
            Indicates valid common prefixes, if a prefix is expected.
        :type prefix_set: set
        :param case:
            Indicates whether the locus_tag is expected to be capitalized.
        :type case: bool
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        if check_value is None:
            check_value = self.genome_id

        results = []
        if only_typo:
            pattern = re.compile(check_value.lower())
            search_result = pattern.search(self.locus_tag.lower())
            if search_result == None:
                results.append("The genome ID is missing.")
        else:
            # Expected structure: SEA_TRIXIE_20
            parts = self.locus_tag.split("_")
            if case:
                if self.locus_tag != self.locus_tag.upper():
                    results.append("The capitalization is incorrect.")
            if len(parts) == 3:
                if prefix_set is not None:
                    if parts[0].upper() not in prefix_set:
                        results.append("The prefix is missing.")
                if parts[1].upper() != check_value.upper():
                    results.append("The genome ID is missing.")
                if not parts[2].isdigit():
                    results.append("The feature number is missing.")
            else:
                results.append("The locus_tag does not contain three parts.")
        result = f"The locus_tag qualifier is {self.locus_tag}. It is "
        if len(results) == 0:
            result = result + "structured correctly."
            status = success
        else:
            result = result + "not structured correctly. " + " ".join(results)
            status = fail
        definition = f"Check if the locus_tag qualifier is structured correctly for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO: create base feature class - fully equivalent to the version in Cds,
    #  but exception clause is appropriately narrow.
    def check_gene_structure(self, eval_id=None, success="correct",
                             fail="error", eval_def=None):
        """
        Check that the gene qualifier contains an integer.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        try:
            value = int(self.gene)
        except ValueError:
            value = self.gene

        result = f"The gene qualifier is {self.gene}. It "
        if isinstance(value, int):
            result += "contains an integer, as expected."
            status = success
        else:
            result += "does not contain an integer, which is not expected."
            status = fail

        definition = f"Check if the gene qualifier contains an integer for " \
                     f"{self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO: create base feature class
    def check_compatible_gene_and_locus_tag(self, eval_id=None,
                                            success="correct", fail="error",
                                            eval_def=None):
        """
        Check that gene and locus_tag attributes contain identical numbers
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The numbers in the gene ({self.gene}) and " \
                 f"locus_tag ({self.locus_tag}) qualifiers are "
        if self.gene == self._locus_tag_num:
            result += "consistent."
            status = success
        else:
            result += "not consistent."
            status = fail

        definition = f"Check if the gene and locus_tag numbers are " \
                     f"consistent for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_peptide_tag_valid(self, eval_id=None, success="correct",
                                fail="error", eval_def=None):
        """
        Checks whether the annotated peptide tag contains any letters
        not strictly within the protein alphabet.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The annotated peptide tag '{self.peptide_tag}' is "
        if self.peptide_tag != "" and self.peptide_tag[-1] == "*":
            letters = set(self.peptide_tag[:-1])
            bad_letters = letters.difference(IUPAC.IUPACProtein.letters)
            if not bad_letters:
                result += "formatted correctly, and uses the correct " \
                          "alphabet."
                status = success
            else:
                result += f"formatted correctly, but has " \
                          f"{len(bad_letters)} characters that fall outside " \
                          f"the protein alphabet."
                status = fail
        elif self.peptide_tag != "" and self.peptide_tag[-1] != "*":
            letters = set(self.peptide_tag)
            bad_letters = letters.difference(IUPAC.IUPACProtein.letters)
            if not bad_letters:
                result += "using the correct alphabet, but doesn't end in '*'."
                status = fail
            else:
                result += f"using {len(bad_letters)} characters outside the " \
                          f"protein alphabet, and doesn't end in '*'."
                status = fail
        else:
            result += "does not exist, so could not be checked."
            status = "unchecked"

        definition = f"Check that the peptide tag appears to be structured " \
                     f"correctly and fits the protein alphabet for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_peptide_tag_correct(self, eval_id=None, success="correct",
                                  fail="error", eval_def=None):
        """
        Checks whether the annotated peptide tag matches the Aragorn
        output.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The annotated peptide tag '{self.peptide_tag}' "

        if self.aragorn_run and self.aragorn_data is not None:
            aragorn_tag = self.aragorn_data["PeptideTag"]

            if self.peptide_tag == aragorn_tag:
                result += "matches the Aragorn output."
                status = success
            else:
                result += f"does not match the Aragorn output ({aragorn_tag})."
                status = fail
        elif self.aragorn_run and self.aragorn_data is None:
            result += "does not match the Aragorn output (no tmRNA)."
            status = fail
        else:
            result += "could not be checked, because Aragorn wasn't run."
            status = "unchecked"

        definition = f"Check that the annotated peptide tag matches the " \
                     f"Aragorn output for {self.locus_tag} {(self.id)}."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_parts(self, eval_id=None, success="correct", fail="error",
                    eval_def=None):
        """
        Makes sure only one region exists for this tRNA.
        :param eval_id: unique identifier for the evaluation
        :type eval_id: str
        :param success: status if the outcome is successful
        :type success: str
        :param fail: status if the outcome is unsuccessful
        :type fail: str
        :param eval_def: description of the evaluation
        :type eval_def: str
        :return:
        """
        result = f"The tRNA has '{self.parts}' parts. This is "
        if self.parts == 1:
            result += "expected."
            status = success
        else:
            result += "unexpected."
            status = fail

        definition = f"Make sure there is only one region for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)



"""Pipeline to install a pdm_utils MySQL database."""

import argparse
from datetime import date
import pathlib
import shutil
import sys


from pdm_utils.classes.alchemyhandler import MySQLDatabaseError
from pdm_utils.constants import constants, db_schema_0
from pdm_utils.functions import (basic, configfile, mysqldb, mysqldb_basic,
                                 pipelines_basic, pipeline_shells, url_basic)
from pdm_utils.pipelines import convert_db

DEFAULT_OUTPUT_FOLDER = "/tmp/"
CURRENT_DATE = date.today().strftime("%Y%m%d")
RESULTS_FOLDER = f"{CURRENT_DATE}_get_db"

DEFAULT_SETTINGS = {"url": constants.DB_SERVER}


# TODO test.
def main(unparsed_args_list):
    """
    Run the get_db pipeline.

    The database data can be retrieved from three places:
    The server, which needs to be downloaded to a new folder.
    A local file, in which no download and no new folder are needed.
    The empty schema stored within pdm_utils, in which no download, new folder,
    or local file are needed.

    :param unparsed_args_list: list of arguments to run the pipeline unparsed
    :type unparsed_args_list: list

    """
    args = parse_args(unparsed_args_list)

    # Set values that are shared between all three options.
    config = configfile.build_complete_config(args.config_file)
    alchemist = pipelines_basic.build_alchemist(None, config=config,
                                                ask_database=False)

    if args.option == "file":
        execute_get_file_db(alchemist, args.database, args.filename,
                            config_file=args.config_file,
                            schema_version=args.schema_version,
                            verbose=args.verbose)
    elif args.option == "new":
        execute_get_new_db(alchemist, args.database, args.schema_version,
                           config_file=args.config_file, verbose=args.verbose)
    else:
        url = args.url
        if url is None:
            server_creds = config["download_server"]
            url = server_creds.get("url")

        if url is None:
            url = DEFAULT_SETTINGS["url"]

        execute_get_server_db(
                    alchemist, args.database, url,
                    folder_path=args.output_folder, db_name=args.db_name,
                    config_file=args.config_file, verbose=args.verbose,
                    subdirectory=args.remote_directory,
                    download_only=args.download_only,
                    get_version=args.get_version, force_pull=args.force_pull,
                    schema_version=args.schema_version)


# TODO test.
def parse_args(unparsed_args_list):
    """
    Verify the correct arguments are selected for getting a new database.
    :param unparsed_args_list: arguments in sys.argv format
    :type unparsed_args_list: list
    :returns: A parsed list of arguments
    :rtype: argparse.Namespace
    """
    get_db_help = ("Pipeline to retrieve and install a new version of a MySQL "
                   "database.")
    database_help = "Name of the MySQL database."
    option_help = "Source of data to create database."
    server_help = "Download database from server."
    url_help = "Server URL from which to retrieve files."
    VERBOSE_HELP = """
        Toggles get_db pipeline progress print statements.
        """
    REMOTE_DIRECTORY_HELP = """
        Remote directory at the server URL from which to retrieve files.
        """
    DB_NAME_HELP = """
        MySQL file option to allow renaming of the new database.
            Follow selection argument with the name of the desired database.
        """
    FORCE_PULL_HELP = """
        Get_db option to force a database pull and ignore database version
        concurrency.
        """
    get_version_help = "Indicates that a .version file should be downloaded."
    output_folder_help = f"Path to the folder to create the folder for " \
                         f"downloading the database. Default is " \
                         f"{DEFAULT_OUTPUT_FOLDER}"
    download_only_help = "The database should be downloaded but not " \
                         "installed locally."
    file_help = "Install database from a SQL file."
    filename_help = "Name of the SQL file and version file."
    new_help = "Create a new empty database."
    schema_version_help = "Database schema version to which the database " \
                          "should be converted."
    config_file_help = "Path to the file containing user-specific login data."

    # database optional for subparser a and required for b and c

    parser = argparse.ArgumentParser(description=get_db_help)

    subparsers = parser.add_subparsers(dest="option", required=True,
                                       help=option_help)

    # Command line structure
    # python3 -m pdm_utils get_db server -db <database>
    parser_a = subparsers.add_parser("server", help=server_help)
    parser_a.add_argument("-db", "--database", type=str, help=database_help,
                          default=None)
    parser_a.add_argument("-u", "--url", type=str,
                          default=None, help=url_help)
    parser_a.add_argument("-gv", "--get_version", action="store_true",
                          default=False, help=get_version_help)
    parser_a.add_argument("-fp", "--force_pull", action="store_true",
                          help=FORCE_PULL_HELP)
    parser_a.add_argument("-o", "--output_folder", type=pathlib.Path,
                          default=None, help=output_folder_help)
    parser_a.add_argument("-d", "--download_only", action="store_true",
                          default=False, help=download_only_help)
    parser_a.add_argument("-n", "--db_name", help=DB_NAME_HELP)
    parser_a.add_argument("-rd", "--remote_directory", type=pathlib.Path,
                          help=REMOTE_DIRECTORY_HELP, default=None)

    parser_b = subparsers.add_parser("file", help=file_help)
    parser_b.add_argument("database", type=str, help=database_help)
    parser_b.add_argument("filename", type=pathlib.Path, help=filename_help)

    parser_c = subparsers.add_parser("new", help=new_help)
    parser_c.add_argument("database", type=str, help=database_help)

    # Add config file option to all subparsers.
    # It could be added after database, but then the optional argument is
    # required to be placed before the required subparser option, which
    # doesn't make sense.
    for p in [parser_a, parser_b, parser_c]:
        p.add_argument("-c", "--config_file", type=pathlib.Path,
                       help=config_file_help, default=None)
        p.add_argument("-v", "--verbose", action="store_true",
                       help=VERBOSE_HELP)
        p.add_argument("-s", "--schema_version", type=int,
                       choices=list(convert_db.CHOICES),
                       default=convert_db.CURRENT_VERSION,
                       help=schema_version_help)

    # Assumed command line arg structure:
    # python3 -m pdm_utils.run <pipeline> <additional args...>
    # sys.argv:      [0]            [1]         [2...]
    args = parser.parse_args(unparsed_args_list[2:])
    return args


def execute_get_server_db(
                alchemist, database, url, folder_path=None,
                folder_name=RESULTS_FOLDER, db_name=None,
                config_file=None, verbose=False, subdirectory=None,
                download_only=False, get_fastas=False, get_alns=False,
                force_pull=False, get_version=False, schema_version=None):
    

    if subdirectory:
        url = "".join([url, str(subdirectory), "/"])

    pool = url_basic.create_pool(pipeline=True)
    if database is None:
        print("Loading get_db interactive environment...")
        cmd = pipeline_shells.GetDBCMD(url, name=alchemist.username, pool=pool)
        cmd.cmdloop(intro=pipeline_shells.GET_DB_CMD_INTRO)

        if cmd.selected is None:
            return

        database = cmd.selected.name
        pkg_url = "".join([cmd.selected.get_abs_path(), "/"])
    else:
        response = url_basic.pool_request(url, pool=pool, pipeline=True)
        directory_listing = url_basic.get_url_listing_dirs(response)

        if database not in directory_listing:
            print("Requested database is not at the specified url.\n"
                  "Please check the database availability.")
            return

        response.close()
        pkg_url = "".join([url, database, "/"])

    if db_name is None:
        db_name = database

    pkg_response = url_basic.pool_request(pkg_url, pool=pool, pipeline=True)

    sql_file_listing = url_basic.get_url_listing_files(pkg_response, "sql")
    version_file_listing = url_basic.get_url_listing_files(
                                                       pkg_response, "version")

    if not sql_file_listing:
        print("Requested database file package does not have a SQL file.\n"
              "Please check SQL file availability at the specified url.")
        return
    database_filename = sql_file_listing[0]

    if not version_file_listing:
        if get_version:
            print("Requested database file package does not have"
                  "a version file.\nPlease check version file availability "
                  "at the specified url.")
            return
        else:
            version_filename = None
    else:
        version_filename = version_file_listing[0]

    if folder_path is None:
        output_path = pipelines_basic.create_working_path(
                                    pathlib.Path(DEFAULT_OUTPUT_FOLDER),
                                    folder_name)
        if output_path.is_dir():
            shutil.rmtree(output_path)
        pipelines_basic.create_working_dir(output_path, force=True)
    else:
        output_path = pipelines_basic.create_working_path(
                                    folder_path, folder_name)
        pipelines_basic.create_working_dir(output_path)

    # Only look for version file is selected.
    if version_filename is not None:
        version_filepath, status1 = prepare_download(
                                            output_path, pkg_url,
                                            version_filename,
                                            "version")
        version_filehandle = version_filepath.open(mode="r")
        version = int(version_filehandle.readline().rstrip())
    else:
        status1 = True
        version = 0

    if (not force_pull) and (version > 0):
        if db_name in alchemist.databases:
            alchemist.database = db_name
            alchemist.build_engine()

            curr_schema_version = mysqldb.get_schema_version(alchemist.engine)
            if curr_schema_version > 2:
                curr_version_data = mysqldb_basic.get_first_row_data(
                                                alchemist.engine, "version")
                curr_version = int(curr_version_data.get("Version", 0))
                if curr_version >= version:
                    print(
                      f"Current database version of {db_name} "
                      "is greater than or equal to the database version "
                      "at the specified listing.\nPlease use "
                      "the --force_pull flag if you would like to "
                      "indiscriminately pull and install a database.")
                    return

    db_filepath, status2 = prepare_download(
                                    output_path, pkg_url, database_filename,
                                    "sql", verbose=verbose)
    if not status1 or not status2:
        print("Unable to download data from server.\n Aborting pipeline.")
        return

    # If downloading from server, user may have selected to not
    # install the database file.
    if (not download_only) and (not get_fastas) and (not get_alns):
        install_db(alchemist, db_name, db_filepath=db_filepath,
                   config_file=config_file, schema_version=schema_version,
                   verbose=verbose)

        # The output folder was only created for downloading from server.
        print("Removing downloaded data.")
        shutil.rmtree(output_path)


def execute_get_file_db(alchemist, database, filename, config_file=None,
                        schema_version=None, verbose=False):
    db_filepath = basic.set_path(filename, kind="file", expect=True)

    install_db(alchemist, database, db_filepath=db_filepath,
               config_file=config_file, schema_version=schema_version,
               verbose=verbose)


def execute_get_new_db(alchemist, database, schema_version,
                       config_file=None, verbose=False):
    install_db(alchemist, database, schema_version=schema_version,
               config_file=config_file, verbose=verbose)


# TODO test.
# TODO move,
def install_db(alchemist, database, db_filepath=None, config_file=None,
               schema_version=None, verbose=False, pipeline=False):
    """
    Install database. If database already exists, it is first removed.
    :param database: Name of the database to be installed
    :type database: str
    :param db_filepath: Directory for installation
    :type db_filepath: Path
    """
    # No need to specify database yet, since it needs to first check if the
    # database exists.
    engine = alchemist.engine
    result = mysqldb_basic.drop_create_db(engine, database)
    engine.dispose()

    if result != 0:
        if pipeline:
            print("Unable to create new, empty database.\nPlease "
                  "check SQL service status and/or database availability.")
            sys.exit(1)

        raise OSError("Unable to create new, empty database.\nPlease "
                      "check SQL service status and/or database availability.")

    alchemist.database = database
    try:
        alchemist.validate_database()
    except MySQLDatabaseError:
        if pipeline:
            print(f"No connection to database {database} due "
                  "to invalid credentials and/or database.")
            sys.exit(1)

        raise MySQLDatabaseError(f"No connection to database {database} due "
                                 "to invalid credentials and/or database.")

    alchemist.build_engine()
    engine = alchemist.engine

    if db_filepath is not None:
        mysqldb_basic.install_db(engine, db_filepath)
    else:
        mysqldb.execute_transaction(engine, db_schema_0.STATEMENTS)

    if schema_version is not None:
        curr_schema_version = mysqldb.get_schema_version(engine)

        if not curr_schema_version == schema_version:
            if verbose:
                print(f"Schema version {curr_schema_version} "
                      "database detected.\nBeginning database conversion to "
                      f"schema version {schema_version}...")
            convert_args = ["pdm_utils.run", "convert", database,
                            "-s", str(schema_version)]
            if verbose:
                convert_args.append("-v")

            if config_file is not None:
                convert_args.extend(["-c", config_file])

            convert_db.main(convert_args)

    engine.dispose()


# TODO test.
def prepare_download(local_folder, url_folder, db_name, extension,
                     verbose=False):
    """
    Construct filepath and check if it already exists, then download.
    :param local_folder: Working directory where the database is downloaded
    :type local_folder: Path
    :param url_folder: Base url where db_files are located.
    :type url_folder: str
    :param db_name: Name of the database to be downloaded
    :type db_name: str
    :param extension: file extension for the database
    :type extension: str
    :returns: Path to the destination directory and the status of the download
    :rtype: Path, bool
    """
    filename = ".".join([db_name, extension])
    url_path = url_folder + filename
    local_path = pathlib.Path(local_folder, filename)
    if local_path.exists():
        print(f"The file {filename} already exists.")
        status = False
    else:
        if verbose:
            print(" ".join(["Preparing to download:", str(db_name)]))
        status = url_basic.download_file(url_path, local_path, verbose=verbose)
    return local_path, status



"""Represents a structure to pair two Genome objects and
perform comparisons between them to identify inconsistencies."""

from pdm_utils.classes import evaluation
from pdm_utils.functions import basic

# TODO methods should be added to match perfect/imperfect features between the
# two genomes as in GenomeTriad.

class GenomePair:

    # Initialize all attributes:
    def __init__(self):

        self.genome1 = None
        self.genome2 = None
        self.evaluations = []


    # Evaluations
    def set_eval(self, eval_id, definition, result, status):
        """Constructs and adds an Evaluation object to the evaluations list.

        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param definition: Description of the evaluation.
        :type definition: str
        :param result: Description of the outcome of the evaluation.
        :type result: str
        :param status: Outcome of the evaluation.
        :type status: str
        """
        evl = evaluation.Evaluation(eval_id, definition, result, status)
        self.evaluations.append(evl)


    def compare_attribute(self, attribute, expect_same=False, eval_id=None,
                          success="correct", fail="error", eval_def=None):
        """Compare values of the specified attribute in each genome.

        :param attribute: Name of the GenomePair object attribute to evaluate.
        :type attribute: str
        :param expect_same:
            Indicates whether the two attribute values are expected to be
            the same.
        :type expect_same: bool
        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param success: Default status if the outcome is a success.
        :type success: str
        :param fail: Default status if the outcome is not a success.
        :type fail: str
        :param eval_def: Description of the evaluation.
        :type eval_def: str
        """
        try:
            test = True
            value1 = getattr(self.genome1, attribute)
            value2 = getattr(self.genome2, attribute)
        except:
            test = False
            value1 = None
            value2 = None
        if test:

            if value1 == value2:
                actual_same = True
            else:
                actual_same = False

            v1_short = basic.truncate_value(str(value1), 30, "...")
            v2_short = basic.truncate_value(str(value2), 30, "...")
            result = (f"The first genome is ID: {self.genome1.id}, "
                      f"Type: {self.genome1.type}. The '{attribute}' attribute "
                      f" contains: '{v1_short}'. "
                      f"The second genome is ID: {self.genome2.id}, "
                      f"Type: {self.genome2.type}. The '{attribute}' attribute "
                      f" contains: '{v2_short}'. These two values are ")

            if actual_same:
                result = result + "identical, "
            else:
                result = result + "different, "

            if actual_same and expect_same:
                result = result + "as expected."
                status = success
            elif not actual_same and not expect_same:
                result = result + "as expected."
                status = success
            else:
                result = result + "which is not expected."
                status = fail
        else:
            result = f"'{attribute}' is not a valid field to be compared."
            status = "untested"
        definition = f"Compare values of the '{attribute}' attribute in each genome."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def compare_date(self, expect, eval_id=None, success="correct",
                     fail="error", eval_def=None):
        """Compare the date of each genome.

        :param expect:
            Is the first genome expected to be "newer", "equal", or "older"
            than the second genome.
        :type expect: str
        :param eval_id: same as for compare_attribute().
        :param success: same as for compare_attribute().
        :param fail: same as for compare_attribute().
        :param eval_def: same as for compare_attribute().
        """

        if expect in set(["newer", "equal", "older"]):
            if self.genome1.date > self.genome2.date:
                actual = "newer"
                actual2 = actual + "than"
            elif self.genome1.date == self.genome2.date:
                actual = "equal"
                actual2 = actual + "to"
            else:
                actual = "older"
                actual2 = actual + "than"

            msg = (f"The query genome '{self.genome1.id}' date "
                   f"is '{self.genome1.date}'."
                   f"The reference genome '{self.genome2.id}' date "
                   f"is '{self.genome2.date}'."
                   f"The date of query genome is {actual2} the "
                   "date of the reference genome, which is ")

            if actual == expect:
                result = msg + "expected."
                status = success
            else:
                result = msg + "not expected."
                status = fail
        else:
            result = f"'{expect}' is an invalid comparison."
            status = "untested"
        definition = "Compare the date of both genomes."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)



import sqlalchemy
from sqlalchemy.orm import mapper
from sqlalchemy.orm import relationship

from pdm_utils.classes.genome import Genome
from pdm_utils.classes.cds import Cds
from pdm_utils.functions import parsing

#-----------------------------------------------------------------------------
#AUTOMAPPER FUNCTIONS
def get_map(mapper, table):
    """Get SQLAlchemy ORM map object.

    :param mapper: Connected and prepared SQLAlchemy automap base object.
    :type mapper: DeclarativeMeta
    :param table: Case-insensitive table to retrieve a ORM map for.
    :type table: str
    :returns: SQLAlchemy mapped object.
    :rtype: DeclarativeMeta
    """
    table = parsing.translate_table(mapper.metadata, table)
    return mapper.classes[table]

#-----------------------------------------------------------------------------
#CLASSICAL MAPPING FUNCTIONS

def map_cds(metadata):
    gene_table = metadata.tables["gene"]

    cds_properties = {"id" : gene_table.c.GeneID,
                      "genome_id" : gene_table.c.PhageID,
                      "start" : gene_table.c.Start,
                      "stop"  : gene_table.c.Stop,
                      "name"  : gene_table.c.Name,
                      "translation": gene_table.c.Translation,
                      "orientation" : gene_table.c.Orientation,
                      "description" : gene_table.c.Notes,
                      "locus_tag"   : gene_table.c.LocusTag}

    cds_exclusions = [gene_table.c.DomainStatus, gene_table.c.ID]

    map = mapper(Cds, gene_table, properties=cds_properties,
                 exclude_properties=cds_exclusions)

    return map

def map_genome(metadata):
    phage_table = metadata.tables["phage"]
    map_cds(metadata)
   
    genome_properties = {"id"                 : phage_table.c.PhageID,
                         "accession"          : phage_table.c.Accession,
                         "name"               : phage_table.c.Name,
                         "host_genus"         : phage_table.c.HostStrain,
                         "sequence"           : phage_table.c.Sequence,
                         "length"             : phage_table.c.SequenceLength,
                         "date"               : phage_table.c.DateLastModified,
                         "gc"                 : phage_table.c.GC,
                         "annotation_status"  : phage_table.c.Status,
                         "retrieve_record"    : phage_table.c.RetrieveRecord,
                         "annotation_author"  : phage_table.c.AnnotationAuthor,
                         "cluster"            : phage_table.c.Cluster2,
                         "subcluster"         : phage_table.c.Subcluster2,
                         "cds_features"       : relationship(Cds)}

    genome_exclusions = [phage_table.c.Cluster, phage_table.c.Notes]

    map = mapper(Genome, phage_table, properties=genome_properties,
                 exclude_properties=genome_exclusions)

    return map





"""Pipeline to freeze a database."""

import argparse
import pathlib
import sys

from pdm_utils.functions import basic
from pdm_utils.functions import configfile
from pdm_utils.functions import mysqldb, mysqldb_basic
from pdm_utils.functions import parsing
from pdm_utils.functions import querying
from pdm_utils.classes.alchemyhandler import AlchemyHandler
from pdm_utils.classes.filter import Filter

RESET_VERSION = "UPDATE version SET Version = 0"
TARGET_TABLE = "phage"

# TODO unittest.
def main(unparsed_args_list):
    """Run main freeze database pipeline."""
    args = parse_args(unparsed_args_list)
    ref_database = args.database
    reset = args.reset
    new_database = args.new_database_name
    prefix = args.prefix

    # Filters input: phage.Status=draft AND phage.HostGenus=Mycobacterium
    # Args structure: [['phage.Status=draft'], ['phage.HostGenus=Mycobacterium']]
    filters = args.filters

    # Create config object with data obtained from file and/or defaults.
    config = configfile.build_complete_config(args.config_file)
    mysql_creds = config["mysql"]

    # Verify database connection and schema compatibility.
    print("Connecting to the MySQL database...")
    alchemist1 = AlchemyHandler(database=ref_database,
                                username=mysql_creds["user"],
                                password=mysql_creds["password"])
    alchemist1.connect(pipeline=True)
    engine1 = alchemist1.engine
    mysqldb.check_schema_compatibility(engine1, "the freeze pipeline")

    # Get SQLAlchemy metadata Table object
    # table_obj.primary_key.columns is a
    # SQLAlchemy ColumnCollection iterable object
    # Set primary key = 'phage.PhageID'
    alchemist1.build_metadata()
    table = querying.get_table(alchemist1.metadata, TARGET_TABLE)
    for column in table.primary_key.columns:
        primary_key = column

    # Create filter object and then add command line filter strings
    db_filter = Filter(alchemist=alchemist1, key=primary_key)
    db_filter.values = []

    # Attempt to add filters and exit if needed.
    add_filters(db_filter, filters)

    # Performs the query
    db_filter.update()

    # db_filter.values now contains list of PhageIDs that pass the filters.
    # Get the number of genomes that will be retained and build the
    # MYSQL DELETE statement.
    keep_set = set(db_filter.values)
    delete_stmt = construct_delete_stmt(TARGET_TABLE, primary_key, keep_set)
    count_query = construct_count_query(TARGET_TABLE, primary_key, keep_set)
    phage_count = mysqldb_basic.scalar(alchemist1.engine, count_query)

    # Determine the name of the new database.
    if new_database is None:
        if prefix is None:
            prefix = get_prefix()
        new_database = f"{prefix}_{phage_count}"

    # Create the new database, but prevent overwriting of current database.
    if engine1.url.database != new_database:
        result = mysqldb_basic.drop_create_db(engine1, new_database)
    else:
        print("Error: names of the reference and frozen databases are the same.")
        print("No database will be created.")
        result = 1

    # Copy database.
    if result == 0:
        print(f"Reference database: {ref_database}")
        print(f"New database: {new_database}")
        result = mysqldb_basic.copy_db(engine1, new_database)
        if result == 0:
            print(f"Deleting genomes...")
            alchemist2 = AlchemyHandler(database=new_database,
                                        username=engine1.url.username,
                                        password=engine1.url.password)
            alchemist2.connect(pipeline=True)
            engine2 = alchemist2.engine
            engine2.execute(delete_stmt)
            if reset:
                engine2.execute(RESET_VERSION)

            # Close up all connections in the connection pool.
            engine2.dispose()
        else:
            print("Unable to copy the database.")
        # Close up all connections in the connection pool.
        engine1.dispose()
    else:
        print(f"Error creating new database: {new_database}.")
    print("Freeze database script completed.")

# TODO unittest.
def parse_args(unparsed_args_list):
    """Verify the correct arguments are selected."""

    freeze_help = "Pipeline to prepare a database for publication."
    database_help = "Name of the MySQL database."
    filters_help = (
        "Indicates which genomes to retain in the new database, "
        "with each conditional formatted as 'table.Field=value'.")
    new_database_name_help = "The new name of the frozen database"
    prefix_help = "The prefix used in the new name of the frozen database"
    reset_help = "Reset version to 0 in new database."
    config_file_help = "Path to the file containing user-specific login data."

    parser = argparse.ArgumentParser(description=freeze_help)
    parser.add_argument("database", type=str, help=database_help)
    parser.add_argument("-f", "--filters", nargs="?",
                        type=parsing.parse_cmd_string, help=filters_help,
                        default=[])
    parser.add_argument("-n", "--new_database_name", type=str,
        help=new_database_name_help)
    parser.add_argument("-p", "--prefix", type=str,
        help=prefix_help)
    parser.add_argument("-r", "--reset", action="store_true",
        default=False, help=reset_help)
    parser.add_argument("-c", "--config_file", type=pathlib.Path,
                        help=config_file_help, default=None)

    # Assumed command line arg structure:
    # python3 -m pdm_utils.run <pipeline> <additional args...>
    # sys.argv:      [0]            [1]         [2...]
    args = parser.parse_args(unparsed_args_list[2:])
    return args

# TODO unittest.
def get_prefix():
    """Allow user to select appropriate prefix for the new database."""
    OPTIONS = ["none",
               "Actinobacteriophage",
               "Bacteriophage",
               "Cyanobacteriophage",
               "Custom"]
    print("\n\nDefault database prefix options:"
        f"\n1: {OPTIONS[1]}"
        f"\n2: {OPTIONS[2]}"
        f"\n3: {OPTIONS[3]}"
        f"\n4: {OPTIONS[4]}")
    prompt = "Which database prefix should be used? "
    choice = basic.select_option(prompt, {1,2,3,4})
    if choice == 1:
        prefix = OPTIONS[1]
    elif choice == 2:
        prefix = OPTIONS[2]
    elif choice == 3:
        prefix = OPTIONS[3]
    elif choice == 4:
        prefix = input("Provide the custom database prefix: ")
    return prefix

# TODO test.
def construct_set_string(phage_id_set):
    """Convert set of phage_ids to string formatted for MySQL.

    e.g. set: {'Trixie', 'L5', 'D29'}
    returns: "('Trixie', 'L5', 'D29')""
    """
    string = "('" + "', '".join(phage_id_set) + "')"
    return string

# TODO test.
def construct_count_query(table, primary_key, phage_id_set):
    """Construct SQL query to determine count."""
    phage_id_string = construct_set_string(phage_id_set)
    query = (f"SELECT count(*) as count FROM {table} "
             f"WHERE {primary_key} IN {phage_id_string}")
    return query

# TODO test.
def construct_delete_stmt(table, primary_key, phage_id_set):
    """Construct SQL query to determine count."""
    phage_id_string = construct_set_string(phage_id_set)
    statement = (f"DELETE FROM {table} "
                 f"WHERE {primary_key} NOT IN {phage_id_string}")
    return statement

# TODO test.
def add_filters(filter_obj, filters):
    """Add filters from command line to filter object."""
    errors = 0
    for or_filters in filters:
        for filter in or_filters:
            # Catch the error if it is an invalid table.column
            try:
                filter_obj.add(filter)
            except:
                print(f"Invalid filter: {filter}")
                errors += 1
    if errors > 0:
        print("Unable to create new database.")
        sys.exit(1)



"""Pipeline to upgrade or downgrade the schema of a MySQL database."""

import argparse
import pathlib
import sys

from pdm_utils.classes.alchemyhandler import AlchemyHandler
from pdm_utils.functions import configfile
from pdm_utils.functions import mysqldb, mysqldb_basic
from pdm_utils.constants import schema_conversions

MAX_VERSION = 10
CURRENT_VERSION = 10
VERSIONS = list(range(0, MAX_VERSION + 1))
CHOICES = set(VERSIONS)


# TODO unittest.
def get_step_name(dir, step):
    """Generates the name of the script conversion filename."""
    if dir == "upgrade":
        first = step - 1
        second = step
    elif dir == "downgrade":
        first = step + 1
        second = step
    else:
        first = step
        second = step
    step_name = f"{dir}_{first}_to_{second}"
    return step_name


# TODO unittest.
def parse_args(unparsed_args_list):
    """Verify the correct arguments are selected for converting database."""
    convert_help = (
        "Pipeline to upgrade or downgrade the schema of a MySQL database.")
    database_help = "Name of the MySQL database."
    schema_version_help = (
        "Database schema version to which the database should be converted.")
    new_database_name_help = (
        "The new name of the converted database if different from "
        "the original database name.")
    verbose_help = (
        "Conversion progress will be printed.")
    config_file_help = (
        "Path to the file containing user-specific login data.")

    parser = argparse.ArgumentParser(description=convert_help)
    parser.add_argument("database", type=str, help=database_help)
    parser.add_argument("-s", "--schema_version", type=int,
        choices=list(CHOICES), default=CURRENT_VERSION,
        help=schema_version_help)
    parser.add_argument("-n", "--new_database_name", type=str,
        help=new_database_name_help)
    parser.add_argument("-v", "--verbose", action="store_true",
        default=False, help=verbose_help)
    parser.add_argument("-c", "--config_file", type=pathlib.Path,
                        help=config_file_help, default=None)

    # Assumed command line arg structure:
    # python3 -m pdm_utils.run <pipeline> <additional args...>
    # sys.argv:      [0]            [1]         [2...]
    args = parser.parse_args(unparsed_args_list[2:])
    return args


# TODO unittest.
def main(unparsed_args_list):
    """Run main conversion pipeline."""
    # Parse command line arguments
    args = parse_args(unparsed_args_list)
    config = configfile.build_complete_config(args.config_file)
    mysql_creds = config["mysql"]
    alchemist1 = AlchemyHandler(database=args.database,
                                username=mysql_creds["user"],
                                password=mysql_creds["password"])
    alchemist1.connect(pipeline=True)
    engine1 = alchemist1.engine


    target = args.schema_version
    actual = mysqldb.get_schema_version(engine1)
    steps, dir = get_conversion_direction(actual, target)

    # Iterate through list of versions and implement SQL files.
    if dir == "none":
        if args.verbose == True:
            print("No schema conversion is needed.")
        convert = False
    else:
        convert = True

    if convert == True:
        if (args.new_database_name is not None and
                args.new_database_name != args.database):
            result = mysqldb_basic.drop_create_db(engine1, args.new_database_name)
            if result == 0:
                result = mysqldb_basic.copy_db(engine1, args.new_database_name)
                if result == 0:
                    # Create a new connection to the new database.
                    alchemist2 = AlchemyHandler(database=args.new_database_name,
                                                username=engine1.url.username,
                                                password=engine1.url.password)
                    alchemist2.connect(pipeline=True)
                    engine2 = alchemist2.engine

                else:
                    print("Error: Unable to copy the database for conversion.")
                    convert = False
            else:
                print("Error: Unable to create the new database for conversion.")
                convert = False
        else:
            engine2 = engine1

        if convert == True:
            stop_step, summary = convert_schema(engine2, actual, dir,
                                                steps, verbose=args.verbose)
            engine2.dispose()
            if stop_step == target:
                if args.verbose == True:
                    print("\n\nThe database schema conversion was successful.")
            else:
                print("\n\nError: "
                      "The database schema conversion was not successful. "
                      f"Unable to proceed past schema version {stop_step}.")
            if args.verbose == True:
                print_summary(summary)
    engine1.dispose()

# TODO unittest.
def get_conversion_direction(actual, target):
    """Determine needed conversion direction and steps."""
    # Compare actual version to desired selection to get list of versions.
    if actual == target:
        steps = []
        dir = "none"
    elif actual < target:
        dir = "upgrade"
        steps = list(range(actual + 1, target + 1))
    else:
        steps = list(reversed(range(target, actual)))
        dir = "downgrade"
    return steps, dir


# TODO unittest.
def convert_schema(engine, actual, dir, steps, verbose=False):
    """Iterate through conversion steps and convert database schema."""
    summary = {} #Key = conversion step. Value = summary dictionary.
    index = 0
    convert = True
    stop_step = actual
    while (index < len(steps) and convert == True):
        step = steps[index]
        if verbose == True:
            print(f"{dir[:-1].capitalize()}ing to schema version {step}...")
        step_name = get_step_name(dir, step)
        step_dict = get_step_data(step_name)
        commands = step_dict["statements"]
        commands = [i for i in commands if i != ""]
        commands = [i for i in commands if i != "\n"]
        # Try to convert the schema.
        result = mysqldb.execute_transaction(engine, commands)
        if result == 1:
            convert = False
            print("Error encountered while executing MySQL statements.")
        if convert == False:
            print(f"Error: Unable to {dir} schema to version {step}.")
        else:
            stop_step = step
            summary[step_name] = step_dict["step_summary_dict"]
        index += 1
    return stop_step, summary


# TODO unittest.
def get_step_data(step_name):
    """Get dictionary of conversion step data."""
    if step_name not in schema_conversions.CONVERSION_STEPS.keys():
        print(f"Error: The conversion step {step_name} is not available.")
        print("Unable to proceed with database conversion.")
        sys.exit(1)
    else:
        step_dict = schema_conversions.CONVERSION_STEPS[step_name]
        return step_dict


# TODO unittest.
def print_summary(summary):
    """Print summary of data that could be lost or inaccurate."""
    keys = summary.keys()
    if len(keys) > 0:
        for key in keys:
            step = summary[key]
            step_keys = step.keys()
            if len(step_keys) > 0:
                print(f"\n\n\nConversion step: {key}")
            if "LOST_TABLE" in step_keys:
                print("\nWarning: data in the following tables have been lost:")
                for item in step["LOST_TABLE"]:
                    print(item)
            if "LOST_COLUMN" in step_keys:
                print("\nWarning: data in the following columns have been lost:")
                for item in step["LOST_COLUMN"]:
                    print(item)
            if "LOST_DATA" in step_keys:
                print("\nWarning: a subset of data in the "
                      "following columns have been lost:")
                for item in step["LOST_DATA"]:
                    print(item)
            if "INACCURATE_COLUMN" in step_keys:
                print("\nWarning: data in the following columns have "
                       "been auto-populated and may be inaccurate:")
                for item in step["INACCURATE_COLUMN"]:
                    print(item)



"""Misc. functions to manipulate tickets."""

from pdm_utils.constants import constants
from pdm_utils.functions import basic
from pdm_utils.functions import eval_modes
from pdm_utils.classes import ticket
from pdm_utils.classes import genome

def modify_import_data(data_dict, required_keys, optional_keys, keywords):
    """Modifies ticket data to conform to requirements for an ImportTicket object.

    :param data_dict: Dictionary of import ticket data.
    :type data_dict: dict
    :param required_keys: Set of keys required to be in the data dictionary.
    :type required_keys: set
    :param optional_keys:
        Set of optional keys that are not required to be in the data dictionary.
    :type optional_keys: set
    :param keywords:
        Set of valid keyword values that are handled differently
        than other values.
    :type keywords: set
    :returns: Indicates if the ticket is structured properly.
    :rtype: bool
    """
    # Some import table fields are required. Others are optional.
    # First check if all required fields are present.
    missing_req_fields = required_keys - data_dict.keys()
    # Second, check if there are any fields that are not expected.
    extra_fields = data_dict.keys() - required_keys - optional_keys

    # If row in import table has more fields than headers, they are
    # assigned None by csv.DictReader.
    # If row in import table contains fields with no data, they are
    # assigned "" by csv.DictReader.
    # "retrieve" = get data from PhagesDB.
    # "retain" = get data from current genome in the MySQL database.
    # "parse" = get data from the flatfile itself (only for host_genus and accession)
    # "none" = there is no applicable data (only for accession and subcluster)
    # "" = source of data should be automatically determined (optional)
    if (len(missing_req_fields) == 0 and len(extra_fields) == 0):

        # Convert None values to "".
        set_empty(data_dict)
        # Confirm that certain keywords are lowercase.
        set_keywords(data_dict, keywords)
        # Determine which fields are not present, and add them to the dict.
        set_missing_keys(data_dict, optional_keys)
        # Certain fields should be lowercase.
        data_dict["type"] = data_dict["type"].lower()
        data_dict["description_field"] = data_dict["description_field"].lower()
        data_dict["eval_mode"] = data_dict["eval_mode"].lower()
        # For fields that are empty (""), set them with default values.
        set_dict_value(data_dict, "host_genus", "retrieve", "retain")
        set_dict_value(data_dict, "cluster", "retrieve", "retain")
        set_dict_value(data_dict, "subcluster", "retrieve", "retain")
        set_dict_value(data_dict, "annotation_author", "1", "retain")
        set_dict_value(data_dict, "retrieve_record", "1", "retain")
        set_dict_value(data_dict, "annotation_status", "draft", "final")

        # If status = 'draft', accession can remain "".
        if data_dict["annotation_status"] != "draft":
            set_dict_value(data_dict, "accession", "retrieve", "retain")
        return True
    else:
        print(f"The ticket {data_dict} is not formatted correctly.")
        return False



def parse_import_ticket_data(data_dict):
    """Converts import ticket data to a ImportTicket object.

    :param data_dict:
        A dictionary of data with the following keys:

            0. Import action type
            1. Primary PhageID
            2. Host
            3. Cluster
            4. Subcluster
            5. Status
            6. Annotation Author (int)
            7. Feature field
            8. Accession
            9. Retrieve Record (int)
            10. Evaluation mode

    :type data_dict: dict
    :returns: A pdm_utils ImportTicket object.
    :rtype: ImportTicket
    """
    ticket_attributes = constants.IMPORT_TABLE_STRUCTURE["valid_ticket"]
    other_attributes = data_dict.keys() - ticket_attributes

    tkt = ticket.ImportTicket()
    tkt.data_dict = data_dict
    for attr in ticket_attributes:
        attr_value = data_dict[attr]
        setattr(tkt, attr, attr_value)

    data_retrieve = set()
    data_retain = set()
    data_add = set()
    data_parse = set()

    other_attributes = list(other_attributes)
    x = 0
    while x < len(other_attributes):
        attr = other_attributes[x]
        attr_value = data_dict[attr]
        if attr_value == "retrieve":
            data_retrieve.add(attr)
        elif attr_value == "retain":
            data_retain.add(attr)
        elif attr_value == "parse":
            data_parse.add(attr)
        else:
            data_add.add(attr)
        x += 1
    tkt.data_retrieve = data_retrieve
    tkt.data_retain = data_retain
    tkt.data_parse = data_parse
    tkt.data_add = data_add

    return tkt


def set_empty(data_dict):
    """Convert None values to an empty string.

    :param data_dict: Dictionary of import ticket data.
    :type data_dict: dict
    """
    for key in data_dict.keys():
        if data_dict[key] is None:
            data_dict[key] = ""

def set_keywords(data_dict, keywords):
    """Convert specific values in a dictionary to lowercase.

    :param data_dict: Dictionary of import ticket data.
    :type data_dict: dict
    :param keywords:
        Set of valid keyword values that are handled differently
        than other values.
    :type keywords: set
    """
    for key in data_dict.keys():
        value = data_dict[key]
        if isinstance(value, str):
            if value.lower() in keywords:
                data_dict[key] = value.lower()


def set_missing_keys(data_dict, expected_keys):
    """Add a list of keys-values to a dictionary if it doesn't have those keys.

    :param data_dict: Dictionary of import ticket data.
    :type data_dict: dict
    :param expected_keys: Set of keys expected to be in the dictionary.
    :type expected_keys: set
    """
    missing_keys = expected_keys - data_dict.keys()
    for key in missing_keys:
        data_dict[key] = ""

def set_dict_value(data_dict, key, first, second):
    """Set the value for a specific key based on 'type' key-value.

    :param data_dict: Dictionary of import ticket data.
    :type data_dict: dict
    :param key: Dictionary key to change value of.
    :type key: str
    :param first: Value to assign to 'key' if 'type' == 'add'.
    :type first: str
    :param second: Value to assign to 'key' if 'type' != 'add'.
    :type second: str
    """
    if data_dict[key] == "":
        if data_dict["type"] == "add":
            data_dict[key] = first
        else:
            data_dict[key] = second

# TODO it probably makes more sense to pass the 'eval_mode' and
# 'description_field' data into modify_import_data() so that they get added
# to the data_dict. As a result, those values will be exported to the
# revised import table file when the script completes.
def construct_tickets(list_of_data_dict, eval_data_dict, description_field,
                      required_keys, optional_keys, keywords):
    """Construct pdm_utils ImportTickets from parsed data dictionaries.

    :param list_of_data_dict: List of import ticket data dictionaries.
    :type list_of_data_dict: list
    :param eval_data_dict: Dictionary of boolean evaluation flags.
    :type eval_data_dict: dict
    :param description_field:
        Default value to set ticket.description_field attribute if not
        present in the data dictionary.
    :type description_field: str
    :param required_keys: Set of keys required to be in the data dictionary.
    :type required_keys: set
    :param optional_keys:
        Set of optional keys that are not required to be in the data dictionary.
    :type optional_keys: set
    :param keywords:
        Set of valid keyword values that are handled differently
        than other values.
    :type keywords: set
    :returns: List of pdm_utils ImportTicket objects.
    :rtype: list
    """

    tkt_id = 0
    list_of_tickets = []
    for dict in list_of_data_dict:
        tkt_id += 1
        # Each ticket should contain a distinct eval_flag dictionary object.
        input_eval_mode = eval_data_dict["eval_mode"]
        input_eval_flag_dict = eval_data_dict["eval_flag_dict"].copy()

        result = modify_import_data(dict, required_keys, optional_keys, keywords)
        if result:
            tkt = parse_import_ticket_data(dict)
            tkt.id = tkt_id

            # Only set description_field and eval_mode from parameters
            # if they weren't set within the ticket.
            if tkt.description_field == "":
                tkt.description_field = description_field
                # Add description_field to data_dict so that it
                # can be saved to file.
                tkt.data_dict["description_field"] = description_field

            if tkt.eval_mode == "":
                tkt.eval_mode = input_eval_mode
                tkt.eval_flags = input_eval_flag_dict
                # Add eval_mode to data_dict so that it can be saved to file.
                tkt.data_dict["eval_mode"] = input_eval_mode

            else:
                if tkt.eval_mode in eval_modes.EVAL_MODES.keys():
                    tkt.eval_flags = eval_modes.get_eval_flag_dict(tkt.eval_mode)
            list_of_tickets.append(tkt)
        else:
            print("Unable to create ticket.")
    return list_of_tickets



def identify_duplicates(list_of_tickets, null_set=set()):
    """Compare all tickets to each other to identify ticket conflicts.

    Identifies if the same id, PhageID, and Accession
    is present in multiple tickets.

    :param list_of_tickets:
        A list of pdm_utils ImportTicket objects.
    :type list_of_tickets: list
    :param null_set:
        A set of values that may be expected to be duplicated, that
        should not throw errors.
    :type null_set: set
    :returns:
        tuple (tkt_id_dupes, phage_id_dupes)
        WHERE
        tkt_id_dupes(set) is a set of duplicate ticket ids.
        phage_id_dupes(set) is a set of duplicate PhageIDs.
    :rtype: tuple
    """
    tkt_id_list = []
    phage_id_list = []

    # Create separate lists to check each field for duplications.
    # Skip "none" values since they are expected to be duplicated.
    for tkt in list_of_tickets:
        if tkt.id not in null_set:
            tkt_id_list.append(tkt.id)
        if tkt.phage_id not in null_set:
            phage_id_list.append(tkt.phage_id)
    tkt_id_dupe_set = basic.identify_one_list_duplicates(tkt_id_list)
    phage_id_dupe_set = basic.identify_one_list_duplicates(phage_id_list)
    return (tkt_id_dupe_set, phage_id_dupe_set)


def get_genome(tkt, gnm_type=""):
    """Construct a pdm_utils Genome object from a pdm_utils ImportTicket object.

    :param tkt: A pdm_utils ImportTicket object.
    :type tkt: ImportTicket
    :param gnm_type: Identifier for the type of genome.
    :type gnm_type: str
    :returns: A pdm_utils Genome object.
    :rtype: Genome
    """
    gnm = genome.Genome()
    gnm.type = gnm_type
    gnm.set_id(value=tkt.phage_id)
    gnm.name = tkt.phage_id
    if "host_genus" in tkt.data_add:
        gnm.set_host_genus(tkt.data_dict["host_genus"])
    if "accession" in tkt.data_add:
        gnm.set_accession(tkt.data_dict["accession"])
    if "annotation_status" in tkt.data_add:
        gnm.annotation_status = tkt.data_dict["annotation_status"]
    if "cluster" in tkt.data_add:
        gnm.set_cluster(tkt.data_dict["cluster"])
    if "subcluster" in tkt.data_add:
        gnm.set_subcluster(tkt.data_dict["subcluster"])
    if "annotation_author" in tkt.data_add:
        gnm.set_annotation_author(tkt.data_dict["annotation_author"])
    if "retrieve_record" in tkt.data_add:
        gnm.set_retrieve_record(tkt.data_dict["retrieve_record"])
    return gnm



"""Import evaluations descriptions."""
from pdm_utils.constants import constants

def get_string(value_set):
    value_list = list(value_set)
    value_list.sort()
    value_list = [str(x) for x in value_list]
    string = ", ".join(value_list)
    return string

suffix = constants.NAME_SUFFIX
statuses = get_string(constants.ANNOTATION_STATUS_SET)
authors = get_string(constants.ANNOTATION_AUTHOR_SET)
records = get_string(constants.RETRIEVE_RECORD_SET)

pfx1 = "A genome that needs to be added"
pfx2 = "A genome that needs to be replaced"
pfx3 = "The genome is expected to have"
pfx4 = "The genome is not expected to have"
pfx5 = "The new and old genomes are expected to have"
pfx6 = "The bundled data is expected to contain"

sfx1 = "that is already in the database"

EVAL_DESCRIPTIONS = {
    # Target genome to import
    "GNM-EVAL-001": (f"{pfx1} cannot have a PhageID {sfx1} or is ''"),
    "GNM-EVAL-002": (f"{pfx1} cannot have a Name {sfx1} or is ''."),
    "GNM-EVAL-003": (f"{pfx1} cannot have a nucleotide sequence {sfx1} or is ''."),
    "GNM-EVAL-004": (f"{pfx1} is not expected to have a 'final' annotation status."),
    "GNM-EVAL-005": (f"{pfx1} cannot have an Accession {sfx1}."),
    "GNM-EVAL-006": (f"{pfx2} must have a PhageID {sfx1}."),
    "GNM-EVAL-007": (f"{pfx2} must have a nucleotide sequence {sfx1}."),
    "GNM-EVAL-008": (f"{pfx2} is not expected to have a 'draft' annotation status."),
    "GNM-EVAL-009": (f"A genome with 'draft' annotation status is expected to contain the '{suffix}' suffix in the Name."),
    "GNM-EVAL-010": ("A genome with 'draft' annotation status is expected to have 0 CDS descriptions."),
    "GNM-EVAL-011": ("A genome with 'draft' annotation status is expected to have a '' Accession."),
    "GNM-EVAL-012": (f"A genome with 'final' annotation status is not expected to contain the '{suffix}' suffix in the Name."),
    "GNM-EVAL-013": ("A genome with 'final' annotation status is expected to have > 0 CDS descriptions."),
    "GNM-EVAL-014": (f"The genome is not expected to contain the '{suffix}' suffix in the PhageID."),
    "GNM-EVAL-015": (f"{pfx3} a valid Annotation Status: {statuses}."),
    "GNM-EVAL-016": (f"{pfx3} a valid Annotation Author: {authors}."),
    "GNM-EVAL-017": (f"{pfx3} a valid Retrieve Record: {records}."),
    "GNM-EVAL-018": (f"{pfx3} a Cluster that has been previously defined or be a Singleton."),
    "GNM-EVAL-019": (f"{pfx3} no Subcluster or have a Subcluster that has been previously defined."),
    "GNM-EVAL-020": (f"{pfx3} the following Translation Table: 11."),
    "GNM-EVAL-021": (f"{pfx3} a Host Genus that has been previously defined."),
    "GNM-EVAL-022": (f"{pfx3} a Cluster comprised strictly of alphabetic characters (e.g. 'A')."),
    "GNM-EVAL-023": (f"{pfx3} a Subcluster comprised of alphabetic characters followed by numeric characters (e.g. 'A15')."),
    "GNM-EVAL-024": (f"{pfx3} a taxonomically-consistent Cluster and Subcluster. If there is a Subcluster, it should be a member of the Cluster. If the genome is a Singleton, there should be no Subcluster.  "),
    "GNM-EVAL-025": (f"{pfx3} a date recorded."),
    "GNM-EVAL-026": (f"{pfx3} GC content equal to or greater than 0."),
    "GNM-EVAL-027": (f"{pfx3} GC content equal to or less than 100."),
    "GNM-EVAL-028": (f"{pfx3} a length greater than 0."),
    "GNM-EVAL-029": (f"{pfx3} 1 or more annotated CDS features."),
    "GNM-EVAL-030": (f"{pfx3} 0 features that are nested within another feature, that contain the same start and stop coordinates as another feature, or that contain the same stop coordinate as another feature."),
    "GNM-EVAL-031": (f"{pfx4} any ambiguous nucleotides or gaps."),
    "GNM-EVAL-032": (f"{pfx4} any PhageID typos in the Description field."),
    "GNM-EVAL-033": (f"{pfx4} any PhageID typos in the Source field."),
    "GNM-EVAL-034": (f"{pfx4} any PhageID typos in the Organism field."),
    "GNM-EVAL-035": (f"{pfx4} any host genus typos in the Description field."),
    "GNM-EVAL-036": (f"{pfx4} any host genus typos in the Source field."),
    "GNM-EVAL-037": (f"{pfx4} any host genus typos in the Organism field."),
    "GNM-EVAL-038": (f"Since annotation author is True, {pfx3.lower()} one or more specific author(s) listed in the References fields."),
    "GNM-EVAL-039": (f"Since annotation author is True, {pfx4.lower()} to have generic authors (e.g. 'Lastname, Firstname') listed in the References fields."),
    "GNM-EVAL-040": (f"Since annotation author is False, {pfx4.lower()} one or more specific author(s) listed in the References fields."),

    # Current genome in database
    "GNM2-EVAL-001": ("The genome to be replaced is only expected to have a 'draft' annotation status."),

    # Bundled data
    "BNDL-EVAL-001": (f"{pfx6} an import ticket with instructions regarding how to import the genome data."),
    "BNDL-EVAL-002": (f"{pfx6} genome data parsed from a file that will be imported."),
    "BNDL-EVAL-003": (f"{pfx6} genome data parsed from an import ticket that is not present within the file."),
    "BNDL-EVAL-004": (f"{pfx6} genome data parsed from PhagesDB that is not present within the file."),
    "BNDL-EVAL-005": (f"{pfx6} genome data parsed from the MySQL database."),
    "BNDL-EVAL-006": (f"{pfx6} paired genome data from a file and from a MySQL database."),
    "BNDL-EVAL-007": (f"MySQL statements generated from the bundled data are expected to be successfully executed in a MySQL database."),

    # Import ticket
    "TKT-EVAL-001": ("The import ticket is expected to have a unique TicketID."),
    "TKT-EVAL-002": ("The import ticket is expected to have a unique PhageID."),
    "TKT-EVAL-003": ("The import ticket is expected to have a valid type (e.g. Add or Replace)."),
    "TKT-EVAL-004": ("The import ticket is expected to have a valid description field (e.g. Product)."),
    "TKT-EVAL-005": ("The import ticket is expected to have a valid Evaluation Mode (e.g. 'final')."),
    "TKT-EVAL-006": ("The import ticket is expected to have a set of evaluation flags."),
    "TKT-EVAL-007": ("The import ticket is not expected to have an empty PhageID."),
    "TKT-EVAL-008": ("The import ticket is expected to have a compatible type and retain settings."),
    "TKT-EVAL-009": ("The import ticket is expected to have valid add data."),
    "TKT-EVAL-010": ("The import ticket is expected to have valid retain data."),
    "TKT-EVAL-011": ("The import ticket is expected to have valid retrieve data."),
    "TKT-EVAL-012": ("The import ticket is expected to have valid parse data."),

    # Source feature
    "SRC-EVAL-001": (f"{pfx4} any PhageID typos in the Organism qualifier of the Source feature."),
    "SRC-EVAL-002": (f"{pfx4} any host genus typos in the Organism qualifier of the Source feature."),
    "SRC-EVAL-003": (f"{pfx4} any host genus typos in the Host qualifier of the Source feature."),
    "SRC-EVAL-004": (f"{pfx4} any host genus typos in the Lab_Host qualifier of the Source feature."),

    # CDS feature
    "CDS-EVAL-001": ("The CDS feature's translation is not expected to contain any ambiguous amino acids or gaps."),
    "CDS-EVAL-002": ("The CDS feature's nucleotide and amino acid sequence are expected to be consistent."),
    "CDS-EVAL-003": ("The CDS feature is expected to contain a Translation qualifier."),
    "CDS-EVAL-004": ("The CDS feature's Translation_Table qualifier is expected to be 11."),
    "CDS-EVAL-005": ("The CDS feature is expected to have an integer start coordinate > 0."),
    "CDS-EVAL-006": ("The CDS feature's orientation is expected to be either F or R."),
    "CDS-EVAL-007": ("The CDS feature is expected to have a Locus_Tag qualifier."),
    "CDS-EVAL-008": ("The CDS feature is not expected to have any PhageID typos in the Locus_Tag qualifier."),
    "CDS-EVAL-009": ("The CDS feature is expected to have a Gene qualifier."),
    "CDS-EVAL-010": ("The CDS feature's Gene qualifier is expected to be an integer."),
    "CDS-EVAL-011": ("The CDS feature is expected to have the same integer in the Gene and Locus_Tag qualifiers."),
    "CDS-EVAL-012": ("The CDS feature is not expected to have a description in any qualifiers other than the selected qualifier, if the selected qualifier has no description."),
    "CDS-EVAL-013": ("The CDS feature is expected to have an integer stop coordinate > 0."),
    "CDS-EVAL-014": ("The CDS feature is expected to be composed of 1 or more parts."),

    # Paired genomes
    "GP-EVAL-001": (f"{pfx5} the same PhageID."),
    "GP-EVAL-002": (f"{pfx5} the same Nucleotide Sequence."),
    "GP-EVAL-003": (f"{pfx5} the same Length."),
    "GP-EVAL-004": (f"{pfx5} the same Cluster."),
    "GP-EVAL-005": (f"{pfx5} the same Subcluster."),
    "GP-EVAL-006": (f"{pfx5} the same Host Genus."),
    "GP-EVAL-007": (f"{pfx5} the same Annotation Author."),
    "GP-EVAL-008": (f"{pfx5} the same Translation Table."),
    "GP-EVAL-009": (f"{pfx5} the same Retrieve Record."),
    "GP-EVAL-010": (f"The new genome is expected to have a more recent date than the old genome."),
    "GP-EVAL-011": ("The new genome Phage Name is not expected to contain the '_Draft' suffix."),
    "GP-EVAL-012": ("The new genome Annotation Status is not expected to be 'draft'."),
    "GP-EVAL-013": (f"{pfx5} the same Name."),
    "GP-EVAL-014": (f"{pfx5} the same Annotation Status."),
    "GP-EVAL-015": (f"{pfx5} the same Accession."),

    # tRNA feature
    "TRNA-EVAL-001": ("The tRNA feature is expected to have an integer start coordinate > 0."),
    "TRNA-EVAL-002": ("The tRNA feature is expected to have an integer stop coordinate > 0."),
    "TRNA-EVAL-003": ("The tRNA feature is expected to be composed of just 1 part."),
    "TRNA-EVAL-004": ("The tRNA feature's orientation is expected to be either F or R."),
    "TRNA-EVAL-005": ("The tRNA feature is expected to have a Locus Tag qualifier."),
    "TRNA-EVAL-006": ("The tRNA feature is not expected to have any PhageID typos in the Locus Tag qualifier."),
    "TRNA-EVAL-007": ("The tRNA feature is expected to have a Gene qualifier."),
    "TRNA-EVAL-008": ("The tRNA feature's Gene qualifier is expected to be an integer."),
    "TRNA-EVAL-009": ("The tRNA feature is expected to have the same integer in the Gene and Locus Tag qualifiers."),
    "TRNA-EVAL-010": ("The tRNA feature is expected to have a valid isotype (amino acid) prediction."),
    "TRNA-EVAL-011": ("The tRNA feature is expected to have a correct isotype (amino acid) prediction."),
    "TRNA-EVAL-012": ("The tRNA feature is expected to have a valid anticodon prediction."),
    "TRNA-EVAL-013": ("The tRNA feature is expected to have a correct anticodon prediction."),
    "TRNA-EVAL-014": ("The tRNA feature is expected to be ~75bp long (60-100bp allowed)."),
    "TRNA-EVAL-015": ("The tRNA feature is expected to end in CCA, CC, or C."),
    "TRNA-EVAL-016": ("The tRNA feature is expected to have a correct orientation."),
    "TRNA-EVAL-017": ("The tRNA feature is expected to be found by at least one tRNA prediction tool."),
    "TRNA-EVAL-018": ("The tRNA feature is expected to have coordinates that match Aragorn or tRNAscan-SE."),

    # tmRNA feature
    "TMRNA-EVAL-001": ("The tmRNA feature is expected to have an integer start coordinate > 0."),
    "TMRNA-EVAL-002": ("The tmRNA feature is expected to have an integer stop coordinate > 0."),
    "TMRNA-EVAL-003": ("The tmRNA feature is expected to be composed of just 1 part."),
    "TMRNA-EVAL-004": ("The tmRNA feature's orientation is expected to be either F or R."),
    "TMRNA-EVAL-005": ("The tmRNA feature is expected to have a Locus Tag qualifier."),
    "TMRNA-EVAL-006": ("The tmRNA feature is not expected to have any PhageID typos in the Locus Tag qualifier."),
    "TMRNA-EVAL-007": ("The tmRNA feature is expected to have a Gene qualifier."),
    "TMRNA-EVAL-008": ("The tmRNA feature's Gene qualifier is expected to be an integer."),
    "TMRNA-EVAL-009": ("The tmRNA feature is expected to have the same integer in the Gene and Locus Tag qualifiers."),
    "TMRNA-EVAL-010": ("The tmRNA feature is expected to have a peptide tag prediction."),
    "TMRNA-EVAL-011": ("The tmRNA feature is expected to have a correct peptide tag prediction."),
    "TMRNA-EVAL-012": ("The tmRNA feature is expected to have a correct orientation."),
}



"""Configuration file definition and parsing."""

import configparser
import sys

from pdm_utils.functions import basic


def default_sections_keys():
    sections = {"mysql": {"user", "password"},
                "mysqldump": {"user", "password"},
                "ncbi": {"api_key", "email", "tool"},
                "upload_server": {"host", "dest", "user", "password"},
                "download_server": {"url"},
                "emailer": {"username", "password"}}
    return sections


def setup_section(keys, value):
    section = {}
    for key in keys:
        section[key] = value
    return section


def default_parser(null_value):
    """Constructs complete config with empty values."""
    # Need to allow no value if the null value is None.
    null_parser = configparser.ConfigParser(allow_no_value=True)
    all_configs = default_sections_keys()
    for key in all_configs.keys():
        null_parser[key] = setup_section(all_configs[key], null_value)
    return null_parser


def parse_config(file, parser=None):
    """Get parameters from config file."""
    filepath = basic.set_path(file, kind="file", expect=True)
    if parser is None:
        parser = configparser.ConfigParser()

    try:
        parser.read(filepath)
    except:
        print("Unable to parse config file")
        sys.exit(1)
    else:
        return parser


def build_complete_config(file):
    "Buid a complete config object by merging user-supplied and default config."
    parser = default_parser(None)
    if file is not None:
        parser = parse_config(file, parser)
    return parser


def write_config(parser, filepath):
    """Write a ConfigParser to file."""
    with filepath.open("w") as fh:
        parser.write(fh)


def create_empty_config_file(dir, file, null_value):
    """Create an empty config file with all available settings."""
    output_path = basic.set_path(dir, kind="dir", expect=True)
    config_path = basic.make_new_file(output_path, file, "txt", attempt=50)
    if config_path is None:
        print("Unable to create config file. File already exists.")
    else:
        parser = default_parser(null_value)
        write_config(parser, config_path)
